--
-- Copyright 2012 Aaron MacDonald
--
-- Licensed under the Apache License, Version 2.0 (the "License");
-- you may not use this file except in compliance with the License.
-- You may obtain a copy of the License at
--
--     http://www.apache.org/licenses/LICENSE-2.0
--
-- Unless required by applicable law or agreed to in writing, software
-- distributed under the License is distributed on an "AS IS" BASIS,
-- WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
-- implied.
-- See the License for the specific language governing permissions and
-- limitations under the License.
--

local Actor = {}

local Constants = require 'constants'
local Vector = require 'vector'
local Utility = require 'utility'

local Bullet = require 'bullet'
local ColourFlux = require 'colour_flux'

local SoundManager = require 'sound_manager'

local ActorData = require 'data/actors'

----------------------------------------------------------------------

local ACTOR_IMG_PATH = 'actors'

local ACTOR_PLAYER_NAME = 'player'

local MAX_HEIGHT = Constants.TILE_SIZE

local actors_img
local actors_quads

local DAMAGE_RECOVER_TIME = 60

local COLOUR_SPEED = 0.1

local DAMAGE_COLOUR = {255, 0, 0, 255}
local INVINCIBILITY_COLOUR = {128, 128, 255, 255}

local MAX_FREEZE_TIME = 900

local POWER_INVINCIBILITY_SOUND = 'power_invincibility'
local POWER_INVINCIBILITY_FINISH_SOUND = 'power_invincibility_finish'

----------------------------------------------------------------------

local function init_actors_img()
    actors_img = love.graphics.newImage(Utility.img_path(
      ACTOR_IMG_PATH))

    local img_width = actors_img:getWidth()
    local img_height = actors_img:getHeight()

    actors_quads = {}
    for k, v in pairs(ActorData) do
        local sprite = v.sprite

        actors_quads[k] = {}
        actors_quads[k].left = love.graphics.newQuad(
          0, sprite.y,
          sprite.width, sprite.height,
          img_width, img_height
        )
        actors_quads[k].right = love.graphics.newQuad(
          sprite.width, sprite.y,
          sprite.width, sprite.height,
          img_width, img_height
        )
    end
end

local function new(name, map, x, y)
    if not actors_img then
        init_actors_img()
    end

    local sprite_offset_x = 0
    local sprite_offset_y = 0
    local real_y_offset = 0

    local phys_world = map.phys_world

    local body = love.physics.newBody(phys_world, x, y, 'dynamic')

    local width = ActorData[name].sprite.width
    local height = ActorData[name].sprite.height

    local phys_height = height
    if height > MAX_HEIGHT then
        phys_height = MAX_HEIGHT
    end

    local is_vertical = width < phys_height

    local cap_radius
    local cap_inner_width
    local cap_inner_height

    local shapes
    local fixtures = {}

    if is_vertical then
        cap_radius = width / 2
        cap_inner_width = width
        cap_inner_height = phys_height - width

        if cap_inner_height == 0 then
            shapes = {
                love.physics.newCircleShape(
                  0, cap_inner_height / 2, cap_radius
                )
            }
        else
            if cap_inner_height < Constants.MIN_PHYS_SIZE then
                cap_inner_height = Constants.MIN_PHYS_SIZE
            end

            shapes = {
                love.physics.newRectangleShape(
                  0, 0, cap_inner_width, cap_inner_height
                ),

                love.physics.newCircleShape(
                  0, -(cap_inner_height / 2), cap_radius
                ),
                love.physics.newCircleShape(
                  0, cap_inner_height / 2, cap_radius
                )
            }
        end

        sprite_offset_x = cap_radius
        sprite_offset_y = (height - phys_height) + cap_radius
          + (cap_inner_height / 2)

        real_y_offset = (cap_inner_height / 2) + cap_radius
    else
        cap_radius = phys_height / 2
        cap_inner_width = width - phys_height
        cap_inner_height = phys_height

        if cap_inner_width == 0 then
            shapes = {
                love.physics.newCircleShape(
                    cap_inner_width / 2, 0, cap_radius
                )
            }
        else
            if cap_inner_width < Constants.MIN_PHYS_SIZE then
                cap_inner_width = Constants.MIN_PHYS_SIZE
            end

            shapes = {
                love.physics.newRectangleShape(
                  0, 0, cap_inner_width, cap_inner_height
                ),

                love.physics.newCircleShape(
                  -(cap_inner_width / 2), 0, cap_radius
                ),
                love.physics.newCircleShape(
                  cap_inner_width / 2, 0, cap_radius
                )
            }
        end

        sprite_offset_x = width / 2
        sprite_offset_y = (height - phys_height) + cap_radius

        real_y_offset = cap_inner_height / 2
    end

    local shoot_y
    if ActorData[name].force_topmost_shoot_y then
        shoot_y = math.max(sprite_offset_y - (height / 2),
          real_y_offset - Bullet.RADIUS - 1)
    else
        shoot_y = math.min(sprite_offset_y - (height / 2),
          real_y_offset - Bullet.RADIUS - 1)
    end

    for _, s in ipairs(shapes) do
        table.insert(fixtures, love.physics.newFixture(body, s))
    end

    local result = {
        is_player = name == ACTOR_PLAYER_NAME,
        name = name,

        body = body,
        shapes = shapes,
        fixtures = fixtures,

        sprite_offset_x = sprite_offset_x,
        sprite_offset_y = sprite_offset_y,
        real_y_offset = real_y_offset,

        facing_left = true,

        health = ActorData[name].health,

        flies = ActorData[name].flies,

        recover_time = 0,
        invincibility_time = 0,

        recover_flash = ColourFlux.new(
            {DAMAGE_COLOUR, {255, 255, 255, 255}},
            COLOUR_SPEED
        ),
        invincible_flash = ColourFlux.new(
            {INVINCIBILITY_COLOUR, {255, 255, 255, 255}},
            COLOUR_SPEED
        ),

        freeze_time = 0,

        shoot_y = shoot_y,
        shoot_delay = 0
    }

    for _, f in ipairs(fixtures) do
        f:setFriction(0)

        f:setCategory(Constants.PHYS_CAT_ACTOR)
        if ActorData[name].flies then
            f:setMask(Constants.PHYS_CAT_CHASM,
              Constants.PHYS_CAT_ACTOR)
        else
            f:setMask(Constants.PHYS_CAT_ACTOR)
        end

        f:setUserData({
            type = 'actor',
            data = result
        })
    end

    local hitbox = love.physics.newRectangleShape(
      0, (phys_height - height) / 2, width, height
    )

    local hit_fixture = love.physics.newFixture(body, hitbox)
    hit_fixture:setSensor(true)
    hit_fixture:setUserData({
        type = 'hitbox',
        data = result
    })

    if ActorData[name].walk_sound then
        result.walk_sound = SoundManager.new_sound(
            ActorData[name].walk_sound, x, y, nil,
            true, false, true, true
        )
    end

    table.insert(shapes, hitbox)
    table.insert(fixtures, hit_fixture)

    body:setInertia(0)

    return result
end
Actor.new = new

local function make_flyable(act)
    for _, f in ipairs(act.fixtures) do
        f:setMask(Constants.PHYS_CAT_CHASM, Constants.PHYS_CAT_ACTOR)
    end
    act.flies = true

    if act.walk_sound then
        act.walk_sound.source:stop()
        act.walk_sound.persist = false
        act.walk_sound = nil
    end
end
Actor.make_flyable = make_flyable

local function pos(act)
    return Vector.new(
        Utility.round(act.body:getX()),
        Utility.round(act.body:getY())
    )
end
Actor.pos = pos

local function shoot_pos(act)
    local p = pos(act)
    p.y = p.y - act.shoot_y

    return p
end
Actor.shoot_pos = shoot_pos

local function update(act)
    if act.shoot_delay > 0 then
        act.shoot_delay = act.shoot_delay - 1
    end
    if act.freeze_time > 0 then
        act.freeze_time = act.freeze_time - 1
    end

    if act.recover_time > 0 then
        assert(act.invincibility_time == 0)
        act.recover_time = act.recover_time - 1
        if act.recover_time == 0 then
            ColourFlux.reset(act.recover_flash)
        else
            ColourFlux.update(act.recover_flash)
        end
    end

    local p = pos(act)

    if act.walk_sound then
        act.walk_sound.pos = p
    end

    if act.invincibility_time > 0 then
        assert(act.recover_time == 0)
        act.invincibility_time = act.invincibility_time - 1
        if act.invincibility_time == 0 then
            ColourFlux.reset(act.invincible_flash)

            SoundManager.new_sound(
                POWER_INVINCIBILITY_FINISH_SOUND, p.x, p.y
            )
        else
            ColourFlux.update(act.invincible_flash)
        end
    end
end
Actor.update = update

local function can_shoot(act)
    return act.shoot_delay == 0
end
Actor.can_shoot = can_shoot

local function shoot(act, map, angle, supress_sound, triple_shot)
    act.shoot_delay = ActorData[act.name].shoot_delay
    local exploding_shot = act.exploding_shots
      and (act.exploding_shots > 0)

    if not supress_sound then
        local sound

        if exploding_shot then
            if triple_shot then
                sound
                  = ActorData[act.name].explosive_triple_shoot_sound
            else
                sound
                  = ActorData[act.name].explosive_shoot_sound
            end
        else
            if triple_shot then
                sound = ActorData[act.name].triple_shoot_sound
            else
                sound = ActorData[act.name].shoot_sound
            end
        end

        if sound then
            local p = pos(act)
            SoundManager.new_sound(sound, p.x, p.y)
        end
    end

    return Bullet.new(shoot_pos(act), act, angle, map,
      exploding_shot)
end
Actor.shoot = shoot

local function is_alive(act)
    return act.health > 0
end
Actor.is_alive = is_alive

local function hit(act, damage)
    local result = false

    if (act.recover_time == 0) and (act.invincibility_time == 0) then
        result = true

        act.health = act.health - damage

        local sound
        if Actor.is_alive(act) then
            act.recover_time = DAMAGE_RECOVER_TIME
            sound = ActorData[act.name].pain_sound
        else
            if act.walk_sound then
                act.walk_sound.source:stop()
                act.walk_sound.persist = false
            end

            sound = ActorData[act.name].death_sound
        end

        if sound then
            local p = pos(act)
            SoundManager.new_sound(sound, p.x, p.y)
        end
    end

    return result
end
Actor.hit = hit

local function make_invincible(act, frames, supress_sound)
    if act.recover_time > 0 then
        ColourFlux.reset(act.recover_flash)
        act.recover_time = 0
    end
    act.invincibility_time = frames

    if (frames > 0) and not supress_sound then
        local p = pos(act)
        SoundManager.new_sound(POWER_INVINCIBILITY_SOUND, p.x, p.y)
    end
end
Actor.make_invincible = make_invincible

local function freeze(act)
    act.freeze_time = MAX_FREEZE_TIME
end
Actor.freeze = freeze

local function health_percentage(act)
    return act.health / ActorData[act.name].health
end
Actor.health_percentage = health_percentage

local function heal(act, amount)
    local max = ActorData[act.name].health
    act.health = act.health + amount
    if act.health > max then
        act.health = max
    end
end
Actor.heal = heal

local function set_velocity(act, dx, dy, custom_left)
    local body = act.body

    if (dx ~= 0) or (dy ~= 0) then
        if act.walk_sound then
            act.walk_sound.source:play()
        end

        if custom_left ~= nil then
            act.facing_left = custom_left
        else
            if act.facing_left and (dx > 0) then
                act.facing_left = false
            elseif not act.facing_left and (dx < 0) then
                act.facing_left = true
            end
        end

        local speed = ActorData[act.name].speed

        if ((dx * dx) + (dy * dy)) == 2 then
            speed = speed * Constants.DIAGONAL_BASE_SPEED
        end

        local final_dx = dx * speed
        local final_dy = dy * speed

        body:setLinearVelocity(final_dx, final_dy)
    else
        if act.walk_sound then
            act.walk_sound.source:stop()
        end

        if custom_left ~= nil then
            act.facing_left = custom_left
        end

        body:setLinearVelocity(0, 0)
    end
end
Actor.set_velocity = set_velocity

local function draw_data(act)
    local p = pos(act)

    local quad = nil
    if act.facing_left then
        quad = actors_quads[act.name].left
    else
        quad = actors_quads[act.name].right
    end

    local act_colour
    if act.recover_time > 0 then
        act_colour = {ColourFlux.get_colour(act.recover_flash)}
    elseif act.invincibility_time > 0 then
        act_colour = {ColourFlux.get_colour(act.invincible_flash)}
    end

    local result = {
        {
            img = actors_img,
            quad = quad,

            x = p.x - act.sprite_offset_x,
            y = p.y - act.sprite_offset_y,

            real_y = p.y + act.real_y_offset,
            z = 1,

            colour = act_colour
        }
    }

    if act.freeze_time > 0 then
        local _, _, width, height = result[1].quad:getViewport()

        -- Position bottom of rectangle just below sprite to force
        -- rectangle on top
        local real_height
          = (height + 1) * (act.freeze_time / MAX_FREEZE_TIME)

        table.insert(result, {
            is_rectangle = true,

            x = result[1].x,
            y = result[1].y + height - real_height,

            width = width,
            height = real_height,

            real_y = result[1].real_y + 1,
            z = result[1].z,

            colour = {192, 224, 255, 128}
        })
    end

    return result
end
Actor.draw_data = draw_data

----------------------------------------------------------------------

return Actor