--
-- Copyright 2012 Aaron MacDonald
--
-- Licensed under the Apache License, Version 2.0 (the "License");
-- you may not use this file except in compliance with the License.
-- You may obtain a copy of the License at
--
--     http://www.apache.org/licenses/LICENSE-2.0
--
-- Unless required by applicable law or agreed to in writing, software
-- distributed under the License is distributed on an "AS IS" BASIS,
-- WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
-- implied.
-- See the License for the specific language governing permissions and
-- limitations under the License.
--

local AI = {}

local Constants = require 'constants'
local Utility = require 'utility'
local Random = require 'random'

local ScreenManager = require 'screen_manager'

local Vector = require 'vector'
local Graph = require 'graph'
local Grid = require 'grid'

local Dungeon = require 'dungeon'

local Actor = require 'actor'
local ActorData = require 'data/actors'

----------------------------------------------------------------------

local PATH_RECALC_DIST = 256
local PATH_RECALC_DIST_SQR = PATH_RECALC_DIST * PATH_RECALC_DIST

local CHASE_DIST = (Dungeon.TOTAL_ROOM_SIZE * 2)
  * Constants.TILE_SIZE
local CHASE_DIST_SQR = CHASE_DIST * CHASE_DIST

local SHOOT_DIST_SQR
  = ScreenManager.BASE_SCREEN_SIZE * ScreenManager.BASE_SCREEN_SIZE

local OPEN_TILES = {
    [Dungeon.TILES.floor] = true,
    [Dungeon.TILES.treasure] = true,
}
local FLYER_OPEN_TILES = {
    [Dungeon.TILES.chasm] = true
}

----------------------------------------------------------------------

local MAX_PATHFINDS_PER_FRAME = 10
local pathfinds = 0

----------------------------------------------------------------------

local function reset_globals()
    pathfinds = 0
end
AI.reset_globals = reset_globals

----------------------------------------------------------------------

local function tile_vec(actor)
    local v = Actor.pos(actor)
    v.x = math.floor(v.x / Constants.TILE_SIZE)
    v.y = math.floor(v.y / Constants.TILE_SIZE)
    return v
end

local function tile_clear(tile, map_grid, door_tiles, can_fly)
    local tile_type = Grid.get(map_grid, tile.x, tile.y)

    local result = (OPEN_TILES[tile_type]
      or (can_fly and FLYER_OPEN_TILES[tile_type]))
     and not Grid.get(door_tiles, tile.x, tile.y)

    return result
end

local function tile_accessible(tile, start_tile, map_grid, door_tiles,
  can_fly)
    local result = tile_clear(tile, map_grid, door_tiles, can_fly)

    if result then
        local _, dir = Vector.get_dir(start_tile, tile)
        if (dir.x ~= 0) and (dir.y ~= 0) then
            local adj_1 = Vector.add(start_tile, Vector.new(dir.x, 0))
            local adj_2 = Vector.add(start_tile, Vector.new(0, dir.y))

            local r1 = tile_clear(adj_1, map_grid, door_tiles,
              can_fly)
            local r2 = tile_clear(adj_2, map_grid, door_tiles,
              can_fly)

            result = r1 and r2
        end
    end

    return result
end

local function actor_over_chasm(act, map_grid)
    local t = tile_vec(act)
    local tile_type = Grid.get(map_grid, t.x, t.y)
    return tile_type == Dungeon.TILES.chasm
end

-- Bresenham-based supercover line algorithm
-- http://eugen.dedu.free.fr/projects/bresenham/
local function clear_line_of_sight(monster, player, map_grid,
  door_tiles)
    local result = true

    local pv = tile_vec(player)
    local mv = tile_vec(monster)

    local step = Vector.new()
    local error_value = 0
    local error_prev = 0

    local current_p = mv
    local delta = Vector.subtract(pv, current_p)

    if delta.x < 0 then
        step.x = -1
        delta.x = delta.x * -1
    else
        step.x = 1
    end

    if delta.y < 0 then
        step.y = -1
        delta.y = delta.y * -1
    else
        step.y = 1
    end

    local double_delta = Vector.scale(delta, 2)
    if double_delta.x >= double_delta.y then -- First octant
        error_value = delta.x
        error_prev = delta.x

        for i = 0, delta.x - 1 do
            current_p.x = current_p.x + step.x
            error_value = error_value + double_delta.y

            if error_value > double_delta.x then
                current_p.y = current_p.y + step.y
                error_value = error_value - double_delta.x

                if (error_value + error_prev) < double_delta.x then
                    -- Bottom square
                    local tile = Vector.new(
                        current_p.x, current_p.y - step.y
                    )
                    if not tile_clear(tile, map_grid, door_tiles,
                      true) then
                        result = false
                        break
                    end
                elseif (error_value + error_prev) > double_delta.x
                  then
                    -- Left square
                    local tile = Vector.new(
                        current_p.x - step.x, current_p.y
                    )
                    if not tile_clear(tile, map_grid, door_tiles,
                      true) then
                        result = false
                        break
                    end
                else
                    -- Corner (both bottom and left squares)
                    local tile_a = Vector.new(
                        current_p.x, current_p.y - step.y
                    )
                    local tile_b = Vector.new(
                        current_p.x - step.x, current_p.y
                    )
                    if not tile_clear(tile_a, map_grid, door_tiles,
                       true)
                      or not tile_clear(tile_b, map_grid, door_tiles,
                       true) then
                        result = false
                        break
                    end
                end
            end

            -- Current square
            if not tile_clear(current_p, map_grid, door_tiles, true)
              then
                result = false
                break
            end

            error_prev = error_value
        end
    else
        error_value = delta.y
        error_prev = delta.y

        for i = 0, delta.y - 1 do
            current_p.y = current_p.y + step.y
            error_value = error_value + double_delta.x

            if error_value > double_delta.y then
                current_p.x = current_p.x + step.x
                error_value = error_value - double_delta.y

                if (error_value + error_prev) < double_delta.y then
                    local tile = Vector.new(
                        current_p.x - step.x, current_p.y
                    )
                    if not tile_clear(tile, map_grid, door_tiles,
                      true) then
                        result = false
                        break
                    end
                elseif (error_value + error_prev) > double_delta.y
                  then
                    local tile = Vector.new(
                        current_p.x, current_p.y - step.y
                    )
                    if not tile_clear(tile, map_grid, door_tiles,
                      true) then
                        result = false
                        break
                    end
                else
                    local tile_a = Vector.new(
                        current_p.x - step.x, current_p.y
                    )
                    local tile_b = Vector.new(
                        current_p.x, current_p.y - step.y
                    )
                    if not tile_clear(tile_a, map_grid, door_tiles,
                       true)
                      or not tile_clear(tile_b, map_grid, door_tiles,
                       true) then
                        result = false
                        break
                    end
                end
            end

            -- Current square
            if not tile_clear(current_p, map_grid, door_tiles, true)
              then
                result = false
                break
            end

            error_prev = error_value
        end
    end

    return result
end

----------------------------------------------------------------------

local function opposite_dir(dir_a, dir_b)
    return ((dir_a.x == 0) or (dir_a.x == -dir_b.x))
      and ((dir_a.y == 0) or (dir_a.y == -dir_b.y))
end

local function random_adj(monster, map_grid, door_tiles, last_dir)
    local destination
    local dir -- Too much floating point crap caused me to explicitly
              -- state the correct direction instead of derive it from
              -- the destination

    local tv = tile_vec(monster)
    local adj = Vector.adjacent(tv)
    local good_adj = {}

    for _, a in ipairs(adj) do
        if tile_accessible(a, tv, map_grid, door_tiles,
          monster.flies) then
            local _, adj_dir = Vector.get_dir(tv, a)
            local data = {destination = a, dir = adj_dir}

            table.insert(good_adj, data)

            if last_dir then
                if not opposite_dir(adj_dir, last_dir) then
                    table.insert(good_adj, data)
                    table.insert(good_adj, data)
                    table.insert(good_adj, data)
                end
                if Vector.equals(adj_dir, last_dir) then
                    table.insert(good_adj, data)
                    table.insert(good_adj, data)
                    table.insert(good_adj, data)
                    table.insert(good_adj, data)
                end
            end
        end
    end

    if #good_adj > 0 then
        data = Random.select_from_array(good_adj)

        destination = data.destination
        dir = data.dir

        Vector.scale_inline(destination, Constants.TILE_SIZE)
        destination.x = destination.x + Constants.TILE_HALF_SIZE
        destination.y = destination.y + Constants.TILE_HALF_SIZE
    end

    return destination, dir
end

----------------------------------------------------------------------

local function heuristic_cost(start_v, end_v)
    return Vector.distance(start_v, end_v)
end

local function a_star(start_v, end_v, map_grid, can_fly, door_tiles)
    assert(start_v)
    assert(end_v)

    local path = {}

    if pathfinds < MAX_PATHFINDS_PER_FRAME then
        pathfinds = pathfinds + 1
    else
        return {}
    end

    local visited = {}
    local open = { start_v }

    local parent = {}
    Utility.set_double_key(parent, start_v.x, start_v.y, nil)

    local g_score = {}
    Utility.set_double_key(g_score, start_v.x, start_v.y, 0)

    local h_score = {}
    Utility.set_double_key(
        h_score, start_v.x, start_v.y, heuristic_cost(start_v, end_v)
    )

    local f_score = {}
    Utility.set_double_key(
        f_score, start_v.x, start_v.y,
        g_score[start_v.x][start_v.y] + h_score[start_v.x][start_v.y]
    )

    while #open > 0 do
        local vec = open[1]

        if Vector.equals(vec, end_v) then
            break
        end

        table.remove(open, 1)
        Utility.set_double_key(visited, vec.x, vec.y, true)

        local adj = Vector.adjacent(vec)
        for _, a in ipairs(adj) do
            if (tile_accessible(a, vec, map_grid, door_tiles, can_fly)
               or Vector.equals(a, end_v))
              and not (visited[a.x] and visited[a.x][a.y]) then
                local tentative_g_score = g_score[vec.x][vec.y]
                  + Vector.distance(vec, a)
                local tentative_is_better = false

                if not Utility.find_in_array(open, a, Vector.equals)
                  then
                    table.insert(open, a)
                    Utility.set_double_key(
                        h_score, a.x, a.y,
                        heuristic_cost(a, end_v)
                    )
                    tentative_is_better = true
                elseif tentative_g_score < g_score[a.x][a.y] then
                    tentative_is_better = true
                end

                if tentative_is_better then
                    Utility.set_double_key(parent, a.x, a.y, vec)
                    Utility.set_double_key(
                        g_score, a.x, a.y, tentative_g_score
                    )
                    Utility.set_double_key(
                        f_score, a.x, a.y,
                        g_score[a.x][a.y] + h_score[a.x][a.y]
                    )
                end
            end
        end

        table.sort(open, function(a, b)
            return f_score[a.x][a.y] < f_score[b.x][b.y]
        end)
    end

    -- AIs should be trying to find a path to the player only when
    -- it's possible to do so.
    -- Whenever an AI fails to find a path, it repeatedly continues to
    -- do so and ends up freezing the program.  With this assert the
    -- program will just crash, but at least the player won't have to
    -- manually kill the program.
    assert(parent[end_v.x] and parent[end_v.x][end_v.y],
      'AI tried to create an invalid path: '
      .. Vector.to_string(start_v) .. ' to '
      .. Vector.to_string(end_v))

    if parent[end_v.x] and parent[end_v.x][end_v.y] then
        local current_v = parent[end_v.x][end_v.y]

        while current_v and not Vector.equals(start_v, current_v) do
            local destination = Vector.scale(current_v,
              Constants.TILE_SIZE)
            destination.x = destination.x + Constants.TILE_HALF_SIZE
            destination.y = destination.y + Constants.TILE_HALF_SIZE

            local parent_v = parent[current_v.x][current_v.y]
            local dir
            if parent_v then
                _, dir = Vector.get_dir(parent_v, current_v)
            else
                _, dir = Vector.get_dir(start_v, current_v)
            end

            table.insert(path, 1, {
                destination = destination, dir = dir
            })

            current_v = parent_v
        end
    end

    return path
end

----------------------------------------------------------------------

local function calculate_wander_path(monster, map_grid, room_graph,
  door_tiles, unlocked_regions)
    local monster_vec = tile_vec(monster)

    local adj_rooms = Graph.get_adjacent(room_graph,
      monster.current_room)

    local monster_unlocked = unlocked_regions[
      monster.current_room.section
    ]

    local room = Random.remove_from_array(adj_rooms)
    while (monster_unlocked and not unlocked_regions[room.section])
      or (not monster_unlocked
       and (room.section ~= monster.current_room.section)) do
        room = Random.remove_from_array(adj_rooms)
    end

    local min_x = (room.x * Dungeon.TOTAL_ROOM_SIZE)
      + Dungeon.WALL_SIZE
    local min_y = (room.y * Dungeon.TOTAL_ROOM_SIZE)
      + Dungeon.WALL_SIZE
    local max_x = min_x + (Dungeon.FULL_ROOM_HALF_SIZE * 2) - 1
    local max_y = min_y + (Dungeon.FULL_ROOM_HALF_SIZE * 2) - 1

    local tile = Vector.new(
        math.random(min_x, max_x), math.random(min_y, max_y)
    )
    local can_fly = monster.flies
    while Vector.equals(monster_vec, tile) or not tile_clear(tile,
      map_grid, door_tiles, can_fly) do
        tile = Vector.new(
            math.random(min_x, max_x), math.random(min_y, max_y)
        )
    end

    return a_star(monster_vec, tile, map_grid, can_fly, door_tiles)
end

local function calculate_chase_path(monster, map_grid, player,
  door_tiles)
    local path = {}

    local monster_vec = tile_vec(monster)
    local player_vec = tile_vec(player)

    if not Vector.equals(monster_vec, player_vec) then
        path = a_star(
            monster_vec, player_vec,
            map_grid, monster.flies, door_tiles
        )
    end

    return path
end

----------------------------------------------------------------------

local function think_about_wandering(monster, player,
  unlocked_regions, room_graph, map_grid)
    local ai_data = monster.ai_data

    if Actor.is_alive(player) then
        local pos = Actor.pos(monster)
        local player_pos = Actor.pos(player)

        local player_in_room = Vector.equals(monster.current_room,
          player.current_room)
        local player_in_adj_room = Graph.are_adjacent(room_graph,
          monster.current_room, player.current_room)

        if ai_data.wandering then
            local not_in_locked_section
              = unlocked_regions[monster.current_room.section]

            if not_in_locked_section
              and (not actor_over_chasm(player, map_grid)
               or monster.flies)
              and (player_in_room or player_in_adj_room) then
                local dist_sqr = Vector.distance_sqr(pos, player_pos)

                if dist_sqr <= CHASE_DIST_SQR then
                    ai_data.wandering = false
                    ai_data.path = nil
                end
            end
        else -- Chasing player
            local dist_sqr = Vector.distance_sqr(pos, player_pos)

            ai_data.wandering = (dist_sqr > CHASE_DIST_SQR)
              or (actor_over_chasm(player, map_grid)
               and not monster.flies)
              or (not player_in_room and not player_in_adj_room)
            if ai_data.wandering then
                ai_data.path = nil
            end
        end
    else
        if not ai_data.wandering then
            ai_data.path = nil
        end
        ai_data.wandering = true
    end
end

local function wander(monster, map_grid, room_graph,
  door_tiles, unlocked_regions)
    local destination
    local dir

    local ai_data = monster.ai_data

    if not ai_data.path or (#ai_data.path == 0) then
        ai_data.path = calculate_wander_path(monster, map_grid,
          room_graph, door_tiles, unlocked_regions)
    end

    if #ai_data.path > 0 then
        local data = ai_data.path[1]
        table.remove(ai_data.path, 1)

        destination = data.destination
        dir = data.dir
    else
        destination, dir = random_adj(monster, map_grid, door_tiles)
    end

    return destination, dir
end

local function chase_player(monster, player, map_grid, door_tiles)
    local destination
    local dir

    local ai_data = monster.ai_data

    if not ai_data.path or (#ai_data.path == 0) then
        ai_data.path = calculate_chase_path(monster, map_grid, player,
          door_tiles)
    else
        assert(#ai_data.path > 0)

        local end_dest = ai_data.path[#ai_data.path].destination

        local dist_sqr = Vector.distance_sqr(
          end_dest, Actor.pos(player))
        if dist_sqr > PATH_RECALC_DIST_SQR then
            ai_data.path = calculate_chase_path(monster, map_grid,
              player, door_tiles)
        end
    end

    if #ai_data.path > 0 then
        local data = ai_data.path[1]
        table.remove(ai_data.path, 1)

        destination = data.destination
        dir = data.dir
    else
        destination, dir = random_adj(monster, map_grid, door_tiles)
    end

    return destination, dir
end

local function plan_move(monster, pos, destination, dir)
    local dist_per_frame
      = ActorData[monster.name].speed * Constants.SECONDS_PER_FRAME

    local ai_data = monster.ai_data
    ai_data.dir = dir

    local dist = Vector.distance(pos, destination)
    ai_data.move_frames_left = math.floor(dist / dist_per_frame)

    local full_frames_dist = dist_per_frame * ai_data.move_frames_left
    local extra_dist = dist - full_frames_dist

    ai_data.extra_dist = extra_dist / dist_per_frame

    if ai_data.extra_dist > 0 then
        ai_data.move_frames_left = ai_data.move_frames_left + 1
    end
end

local function move(monster)
    local ai_data = monster.ai_data

    local dir_mult = 1
    if (ai_data.extra_dist > 0)
      and (ai_data.move_frames_left == 1) then
        dir_mult = ai_data.extra_dist
    end

    Actor.set_velocity(monster,
      ai_data.dir.x * dir_mult, ai_data.dir.y * dir_mult)

    ai_data.move_frames_left = ai_data.move_frames_left - 1
end

----------------------------------------------------------------------

local function run(monster, map_grid, player, unlocked_regions,
  door_tiles, room_graph, map)
    local ai_data = monster.ai_data

    if not ai_data.move_frames_left
      or ai_data.move_frames_left == 0 then
        think_about_wandering(monster, player, unlocked_regions,
          room_graph, map_grid)

        local destination
        local dir

        if ai_data.wandering then
            destination, dir = wander(monster, map_grid, room_graph,
              door_tiles, unlocked_regions)
        else
            destination, dir = chase_player(monster, player, map_grid,
              door_tiles)
        end

        plan_move(monster, Actor.pos(monster), destination, dir)
    end

    move(monster)

    local new_bullet
    if Actor.is_alive(player) then

        local player_pos = Actor.pos(player)
        local mon_pos = Actor.pos(monster)
        
        if ActorData[monster.name].shoot_delay
          and Actor.can_shoot(monster)
          and (Vector.distance_sqr(mon_pos, player_pos)
           < SHOOT_DIST_SQR)
          and clear_line_of_sight(monster, player, map_grid,
           door_tiles) then
            player_pos.y = player_pos.y + player.shoot_y
            mon_pos.y = mon_pos.y + monster.shoot_y

            local delta = Vector.subtract(player_pos, mon_pos)
            local angle = Utility.coords_to_angle(delta.x, delta.y)

            new_bullet = Actor.shoot(monster, map, angle)
        end
    end

    return new_bullet
end
AI.run = run

----------------------------------------------------------------------

return AI