--
-- Copyright 2012 Aaron MacDonald
--
-- Licensed under the Apache License, Version 2.0 (the "License");
-- you may not use this file except in compliance with the License.
-- You may obtain a copy of the License at
--
--     http://www.apache.org/licenses/LICENSE-2.0
--
-- Unless required by applicable law or agreed to in writing, software
-- distributed under the License is distributed on an "AS IS" BASIS,
-- WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
-- implied.
-- See the License for the specific language governing permissions and
-- limitations under the License.
--

local Bullet = {}

local Constants = require 'constants'
local Utility = require 'utility'

local SoundManager = require 'sound_manager'

local Vector = require 'vector'
local ColourFlux = require 'colour_flux'

local ActorData = require 'data/actors'

----------------------------------------------------------------------

local BULLET_IMG_PATH = 'bullet'

local NORMAL_HIT_WALL_SOUND = 'bullet_hit_wall'
local EXPLODING_HIT_WALL_SOUND = 'exploding_bullet_hit_wall'

local NORMAL_HIT_ACTOR_SOUND = 'bullet_hit_actor'
local EXPLODING_HIT_ACTOR_SOUND = 'exploding_bullet_hit_actor'

local NORMAL_HIT_ACTOR_NO_DAM_SOUND = 'bullet_hit_actor_no_dam'
local EXPLODING_HIT_ACTOR_NO_DAM_SOUND
  = 'exploding_bullet_hit_actor_no_dam'

local RADIUS = 8
Bullet.RADIUS = RADIUS

local SIZE = RADIUS * 2
local BORDER_SIZE = 2

local bullet_img
local bullet_quad

local COLOUR_SPEED = 0.25

local NUM_EXPLOSION_BULLETS = 16
local EXPLOSION_ANGLE = (2 * math.pi) / NUM_EXPLOSION_BULLETS

----------------------------------------------------------------------

local function init_img()
    bullet_img = love.graphics.newImage(Utility.img_path(
      BULLET_IMG_PATH))

    local img_width = bullet_img:getWidth()
    local img_height = bullet_img:getHeight()

    bullet_quad = love.graphics.newQuad(
      0, 0,
      SIZE + BORDER_SIZE, SIZE + BORDER_SIZE,
      img_width, img_height
    )
end

local function new(pos, actor, angle, map, exploding_shot,
  actors_to_ignore)
    if not bullet_img then
        init_img()
    end

    local phys_world = map.phys_world
    local body = love.physics.newBody(phys_world, pos.x, pos.y,
      'dynamic')

    local shape = love.physics.newCircleShape(RADIUS)

    local fixture = love.physics.newFixture(body, shape)
    fixture:setFriction(0)
    fixture:setCategory(Constants.PHYS_CAT_BULLET)
    fixture:setMask(
        Constants.PHYS_CAT_CHASM,
        Constants.PHYS_CAT_ACTOR,
        Constants.PHYS_CAT_BULLET
    )

    body:setInertia(0)

    local result = {
        phys_data = {
            body = body,
            shape = shape,
            fixture = fixture
        },
        actor = actor,

        colour_flux = ColourFlux.new(
            ActorData[actor.name].bullet_colours,
            COLOUR_SPEED
        ),

        exploding_shot = exploding_shot,
        actors_to_ignore = actors_to_ignore
    }

    fixture:setUserData({
        type = 'bullet',
        data = result
    })

    local final_dx
      = math.cos(angle) * ActorData[actor.name].bullet_speed
    local final_dy
      = math.sin(angle) * ActorData[actor.name].bullet_speed

    body:setLinearVelocity(final_dx, final_dy)

    return result
end
Bullet.new = new

local function create_bullet_explosion(pos, actor, map,
  actors_to_ignore)
    local bullets = {}
    for i = 0, NUM_EXPLOSION_BULLETS - 1 do
        table.insert(bullets, new(
            pos, actor, EXPLOSION_ANGLE * i, map, false,
            actors_to_ignore
        ))
    end

    return bullets
end
Bullet.create_bullet_explosion = create_bullet_explosion

local function pos(bullet)
    local body = bullet.phys_data.body
    return Vector.new(
        Utility.round(body:getX()), Utility.round(body:getY())
    )
end
Bullet.pos = pos

local function damage(bullet)
    return ActorData[bullet.actor.name].damage
end
Bullet.damage = damage

local function update(bullet)
    ColourFlux.update(bullet.colour_flux)
end
Bullet.update = update

local function hit_wall(bullet)
    local p = pos(bullet)
    local sound = NORMAL_HIT_WALL_SOUND
    if bullet.exploding_shot then
        sound = EXPLODING_HIT_WALL_SOUND
    end
    SoundManager.new_sound(sound, p.x, p.y)
end
Bullet.hit_wall = hit_wall

local function hit_actor(bullet, damage_done)
    local p = pos(bullet)

    local sound
    if bullet.exploding_shot then
        if damage_done then
            sound = EXPLODING_HIT_ACTOR_SOUND
        else
            sound = EXPLODING_HIT_ACTOR_NO_DAM_SOUND
        end
    else
        if damage_done then
            sound = NORMAL_HIT_ACTOR_SOUND
        else
            sound = NORMAL_HIT_ACTOR_NO_DAM_SOUND
        end
    end
    SoundManager.new_sound(sound, p.x, p.y)
end
Bullet.hit_actor = hit_actor

local function draw_data(bullet)
    local p = pos(bullet)

    return {
        img = bullet_img,
        quad = bullet_quad,

        x = p.x - RADIUS - BORDER_SIZE,
        y = p.y - RADIUS - BORDER_SIZE,

        real_y = p.y + RADIUS + BORDER_SIZE,
        z = 2,

        colour = {ColourFlux.get_colour(bullet.colour_flux)}
    }
end
Bullet.draw_data = draw_data

----------------------------------------------------------------------

return Bullet