--
-- Copyright 2012 Aaron MacDonald
--
-- Licensed under the Apache License, Version 2.0 (the "License");
-- you may not use this file except in compliance with the License.
-- You may obtain a copy of the License at
--
--     http://www.apache.org/licenses/LICENSE-2.0
--
-- Unless required by applicable law or agreed to in writing, software
-- distributed under the License is distributed on an "AS IS" BASIS,
-- WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
-- implied.
-- See the License for the specific language governing permissions and
-- limitations under the License.
--

local Door = {}

local Constants = require 'constants'
local Utility = require 'utility'
local Vector = require 'vector'

----------------------------------------------------------------------

local DOOR_HORIZONTAL_IMG = 'doors_horizontal'
local DOOR_VERTICAL_IMG = 'doors_vertical'

local DOOR_SIZE = 2
Door.DOOR_SIZE = DOOR_SIZE

local GLOW_BORDER = 4

-- Horizontal door

local H_CLOSED_WIDTH = Constants.TILE_SIZE * DOOR_SIZE
local H_CLOSED_HEIGHT = Constants.TILE_SIZE
  + Constants.TILE_HALF_SIZE + (GLOW_BORDER * 2)

local H_CLOSED_TRANS_X = Constants.TILE_SIZE
local H_CLOSED_TRANS_Y = Constants.TILE_SIZE + GLOW_BORDER

local H_CLOSED_REAL_TRANS_Y = Constants.TILE_HALF_SIZE

local H_OPEN_WIDTH = Constants.TILE_SIZE * DOOR_SIZE
local H_OPEN_HEIGHT = Constants.TILE_SIZE + (GLOW_BORDER * 2)

local H_OPEN_REAL_TRANS_Y = Constants.TILE_HALF_SIZE

local H_OPEN_TRANS_X = Constants.TILE_SIZE
local H_OPEN_TRANS_Y = Constants.TILE_HALF_SIZE + GLOW_BORDER

-- Vertical door

local V_CLOSED_WIDTH = Constants.TILE_SIZE + (GLOW_BORDER * 2)
local V_CLOSED_HEIGHT = Constants.TILE_SIZE * DOOR_SIZE

local V_CLOSED_TRANS_X = Constants.TILE_HALF_SIZE + GLOW_BORDER
local V_CLOSED_TRANS_Y = Constants.TILE_SIZE
  + Constants.TILE_HALF_SIZE

local V_CLOSED_REAL_TRANS_Y = Constants.TILE_SIZE

local V_OPEN_WIDTH = Constants.TILE_SIZE + (GLOW_BORDER * 2)
local V_OPEN_HEIGHT = (Constants.TILE_SIZE * DOOR_SIZE)
  - Constants.TILE_HALF_SIZE

local V_OPEN_TRANS_X = Constants.TILE_HALF_SIZE + GLOW_BORDER
local V_OPEN_TRANS_Y = Constants.TILE_SIZE

local V_OPEN_REAL_TRANS_Y = Constants.TILE_SIZE

--

local horizontal_image
local horizontal_quads

local vertical_image
local vertical_quads

----------------------------------------------------------------------

local function init_sheet(is_horizontal)
    local img
    local quads

    local closed_width
    local closed_height

    local open_width
    local open_height

    local img_width
    local img_height

    if is_horizontal then
        horizontal_image = love.graphics.newImage(
          Utility.img_path(DOOR_HORIZONTAL_IMG))
        horizontal_quads = {}

        quads = horizontal_quads

        closed_width = H_CLOSED_WIDTH
        closed_height = H_CLOSED_HEIGHT

        open_width = H_OPEN_WIDTH
        open_height = H_OPEN_HEIGHT

        img_width = horizontal_image:getWidth()
        img_height = horizontal_image:getHeight()

        img = horizontal_image
    else
        vertical_image = love.graphics.newImage(
          Utility.img_path(DOOR_VERTICAL_IMG))
        vertical_quads = {}

        quads = vertical_quads

        closed_width = V_CLOSED_WIDTH
        closed_height = V_CLOSED_HEIGHT

        open_width = V_OPEN_WIDTH
        open_height = V_OPEN_HEIGHT

        img_width = vertical_image:getWidth()
        img_height = vertical_image:getHeight()

        img = vertical_image
    end

    quads.closed = {}
    quads.open = {}

    for i = 0, Constants.MAX_DOORS - 1 do
        table.insert(quads.closed, love.graphics.newQuad(
          closed_width * i, 0,
          closed_width, closed_height,
          img_width, img_height
        ))
        table.insert(quads.open, love.graphics.newQuad(
          open_width * i, closed_height,
          open_width, open_height,
          img_width, img_height
        ))
    end

    return img, quads
end

----------------------------------------------------------------------

local function new(phys_world, index, is_horizontal, pos)
    assert(index <= Constants.MAX_DOORS, index)

    if (is_horizontal and not horizontal_image)
      or (not is_horizontal and not vertical_image) then
        img, quads = init_sheet(is_horizontal)
    end

    local body = love.physics.newBody(phys_world, pos.x, pos.y)

    local width = Constants.TILE_SIZE
    local height = Constants.TILE_SIZE
    if is_horizontal then
        width = width * DOOR_SIZE
    else
        height = width * DOOR_SIZE
    end

    local shape = love.physics.newRectangleShape(width, height)

    local fixture = love.physics.newFixture(body, shape)
    fixture:setFriction(0)

    local result = {
        is_horizontal = is_horizontal,
        open = false,
        index = index,
        pos = pos,
        phys_data = {
            body = body,
            shape = shape,
            fixture = fixture
        }
    }

    fixture:setUserData({
        type = 'door',
        data = result
    })

    return result
end
Door.new = new

local function covered_tiles(door)
    local tiles = {}

    if not door.open then
        local x = math.floor(door.pos.x / Constants.TILE_SIZE)
        local y = math.floor(door.pos.y / Constants.TILE_SIZE)

        tiles = {
            Vector.new(x, y),
            Vector.new(x - 1, y - 1),
            Vector.new(x - 1, y),
            Vector.new(x, y - 1),
        }
    end

    return tiles
end
Door.covered_tiles = covered_tiles

local function draw_data(door)
    local img
    local quads
    local trans_x
    local trans_y
    local real_trans_y
    local z

    if door.is_horizontal then
        img = horizontal_image

        if door.open then
            quads = horizontal_quads.open

            trans_x = H_OPEN_TRANS_X
            trans_y = H_OPEN_TRANS_Y

            real_trans_y = H_OPEN_REAL_TRANS_Y
            z = 0
        else
            quads = horizontal_quads.closed

            trans_x = H_CLOSED_TRANS_X
            trans_y = H_CLOSED_TRANS_Y

            real_trans_y = H_CLOSED_REAL_TRANS_Y
            z = 1
        end
    else
        img = vertical_image

        if door.open then
            quads = vertical_quads.open

            trans_x = V_OPEN_TRANS_X
            trans_y = V_OPEN_TRANS_Y

            real_trans_y = V_CLOSED_REAL_TRANS_Y
            z = 0
        else
            quads = vertical_quads.closed

            trans_x = V_CLOSED_TRANS_X
            trans_y = V_CLOSED_TRANS_Y

            real_trans_y = V_OPEN_REAL_TRANS_Y
            z = 1
        end
    end

    return {
        img = img,
        quad = quads[door.index],
        x = door.pos.x - trans_x,
        y = door.pos.y - trans_y,

        real_y = door.pos.y + real_trans_y,
        z = z
    }
end
Door.draw_data = draw_data

----------------------------------------------------------------------

return Door