--
-- Copyright 2012 Aaron MacDonald
--
-- Licensed under the Apache License, Version 2.0 (the "License");
-- you may not use this file except in compliance with the License.
-- You may obtain a copy of the License at
--
--     http://www.apache.org/licenses/LICENSE-2.0
--
-- Unless required by applicable law or agreed to in writing, software
-- distributed under the License is distributed on an "AS IS" BASIS,
-- WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
-- implied.
-- See the License for the specific language governing permissions and
-- limitations under the License.
--

local Dungeon = {}

local Vector = require 'vector'
local Graph = require 'graph'
local Random = require 'random'
local Grid = require 'grid'
local Utility = require 'utility'
local Map = require 'map'
local Constants = require 'constants'
local Door = require 'door'

----------------------------------------------------------------------

local TILE_KEY_RED = 7
local TILE_KEY_YELLOW = 8
local TILE_KEY_GREEN = 9
local TILE_KEY_BLUE = 10

local TILE_DOOR_RED = 11
local TILE_DOOR_YELLOW = 12
local TILE_DOOR_GREEN = 13
local TILE_DOOR_BLUE = 14

local TILES = {
    wall = 0,
    inner_wall = 1,
    floor = 2,
    chasm = 3,
    entrance = 4,
    boss_entrance = 5,
    treasure = 6,
    spawn = 7
}
Dungeon.TILES = TILES

local LAYER_CLIFF = 0
local LAYER_FLOOR = LAYER_CLIFF + 1
local LAYER_WALL_FRONT = LAYER_FLOOR + 1
local LAYER_WALL_TOP = LAYER_WALL_FRONT + 1

local OFFSET_WALL_FRONT = Vector.new(0, 32)
local OFFSET_WALL_TOP = Vector.new(0, -32)

----------------------------------------------------------------------

local NUM_ROOMS = 24 -- Ideally, this will be equal to or greater than
                     -- MAX_SECTIONS.

-- Dungeons are divided into sections separated by coloured doors.
-- Section X contains the key to Section (X + 1).
-- The section the player starts in is the White section.
local MAX_KEYS = Constants.MAX_DOORS

-- Minimum number of rooms a section must have in addition to its
-- entrance room.
local MIN_SECTION_CHILDREN = 2 -- 3 rooms in total

-- 0 = No dead ends are connected
-- 1 = All dead ends that can be connected are connected
local PERCENT_LOOPS = 0.5

-- 0 = Treasure only found in dead ends
-- 1 = Treasure in every room that doesn't already have a key,
--     entrance, or spawn.
local PERCENT_EXTRA_TREASURE = 0

-- 0 = No corridors
-- 1 = Every empty room is a corridor
local PERCENT_CORRIDOR = 0.5

-- 0 = All non-corridor rooms are self contained
-- 1 = All non-corridor rooms expand to their neighbors
local PERCENT_EXPANDED_ROOM = 0.3

local NUM_CHASM_SEEDS = 2
-- 0 = No chasms
-- 1 = No walls besides the ones that divide sections
local PERCENT_CHASM = 0.2

-- 0 = Rooms will never have interior walls
-- 1 = All rooms will effectively be corridors
local PERCENT_INTERIOR_WALL = 0.25

local WALL_SIZE = 2
local ROOM_HALF_SIZE = 4
local CHASM_PADDING = 1
local HALL_HALF_SIZE = 1 -- Should be less than ROOM_HALF_SIZE
local DOOR_HALF_SIZE = math.floor(Door.DOOR_SIZE / 2)

local FULL_ROOM_HALF_SIZE = ROOM_HALF_SIZE + CHASM_PADDING
local TOTAL_ROOM_SIZE = (FULL_ROOM_HALF_SIZE * 2) + (WALL_SIZE * 2)

Dungeon.WALL_SIZE = WALL_SIZE
Dungeon.FULL_ROOM_HALF_SIZE = FULL_ROOM_HALF_SIZE
Dungeon.TOTAL_ROOM_SIZE = TOTAL_ROOM_SIZE

local ROOM_PATTERN_REFLECTIONS = {
    {h = true, v = false},
    {h = false, v = true},
    {h = true, v = true}
}

-- 0 = No monsters (boring)
-- 1 = Every empty tile gets a monster (yikes!)
local PERCENT_MONSTERS = 0.10

-- Repeated names increase frequency
local MONSTER_TYPES = {
    'gridbug',
    'gridbug',
    'bot',
    'bot',

    'troll',
    'invader'
}

----------------------------------------------------------------------

local function Vector_sort(a, b)
    return (a.x < b.x) or ((a.x == b.x) and (a.y < b.y))
end

local function list_percentage(list, percent)
    return Utility.round(#list * percent)
end

local function coord_to_room_ref(room_graph, coord)
    return Graph.get_vertex_by_value(room_graph, coord, Vector.equals)
end

local function adj_room_ref(room_graph, room, dir)
    return Graph.get_vertex_by_value(
        room_graph,
        Vector.add(room, Vector.DIRS[dir]),
        Vector.equals
    )
end

----------------------------------------------------------------------

-- Generate the graph/tree of rooms and the dungeon's bounding
-- rectangle
local function room_graph()
    local room_graph = Graph.new()

    local num_rooms = NUM_ROOMS

    local min_room = Vector.new()
    local dungeon_bounds = Vector.new()

    local start_room = Vector.new(0, 0)
    start_room.section = 1

    Graph.add_vertex(room_graph, start_room)
    local frontier_rooms = {start_room}

    num_rooms = num_rooms - 1

    while num_rooms > 0 do
        local room, room_i = Random.select_from_array(frontier_rooms)

        local adjacent = Vector.cardinal_adjacent(room)
        local next_room = nil
        while #adjacent > 0 do
            local adj = Random.remove_from_array(adjacent)
            if not coord_to_room_ref(room_graph, adj) then
                next_room = adj
                break
            end
        end

        if next_room then
            next_room.section = 1
            Graph.add_vertex(room_graph, next_room)
            Graph.set_edge(room_graph, room, next_room, {})

            table.insert(frontier_rooms, next_room)

            if next_room.x < min_room.x then
                min_room.x = next_room.x
            end
            if next_room.y < min_room.y then
                min_room.y = next_room.y
            end

            if next_room.x > dungeon_bounds.x then
                dungeon_bounds.x = next_room.x
            end
            if next_room.y > dungeon_bounds.y then
                dungeon_bounds.y = next_room.y
            end

            num_rooms = num_rooms - 1
        else
            table.remove(frontier_rooms, room_i)
        end
    end

    for k, _ in pairs(room_graph.vertices) do
        k.x = k.x - min_room.x
        k.y = k.y - min_room.y
    end

    dungeon_bounds = Vector.subtract(dungeon_bounds, min_room)
    dungeon_bounds.x = dungeon_bounds.x + 1
    dungeon_bounds.y = dungeon_bounds.y + 1

    return room_graph, dungeon_bounds
end

-- Pick a room at the edge of the dungeon to be the entrance
local function set_entrance(room_graph, dungeon_bounds)
    local entrance = nil

    local dir_name = Random.select_from_array(
      {'north', 'east', 'south', 'west'})

    local x
    local y

    if (dir_name == 'north') or (dir_name == 'south') then
        x = math.random(dungeon_bounds.x - 1)
    end
    if (dir_name == 'east') or (dir_name == 'west') then
        y = math.random(dungeon_bounds.y - 1)
    end

    local room

    if dir_name == 'north' then
        for y = 0, dungeon_bounds.y - 1 do
            room = coord_to_room_ref(room_graph, Vector.new(x, y))
            if room then
                entrance = room
                break
            end
        end
    elseif dir_name == 'east' then
        for x = dungeon_bounds.x - 1, 0, -1 do
            room = coord_to_room_ref(room_graph, Vector.new(x, y))
            if room then
                entrance = room
                break
            end
        end
    elseif dir_name == 'south' then
        for y = dungeon_bounds.y - 1, 0, -1 do
            room = coord_to_room_ref(room_graph, Vector.new(x, y))
            if room then
                entrance = room
                break
            end
        end
    elseif dir_name == 'west' then
        for x = 0, dungeon_bounds.x - 1 do
            room = coord_to_room_ref(room_graph, Vector.new(x, y))
            if room then
                entrance = room
                break
            end
        end
    end

    entrance.entrance = true

    return entrance
end

-- Split dungeon into sections that will be separated by locked doors
local function create_sections(room_graph, entrance)
    local room_parents = {}
    local room_children = {}
    local room_children_count = {}

    local dead_ends = {}
    local non_dead_ends = {}

    Graph.depth_first_search(
        room_graph, entrance,
        function (room, parent)
            room_parents[room] = parent

            room_children[room] = {}
            room_children_count[room] = 0

            while parent do
                room_children[parent][room] = true
                room_children_count[parent]
                  = room_children_count[parent] + 1

                parent = room_parents[parent]
            end

            if #Graph.get_adjacent(room_graph, room) == 1 then
                dead_ends[room] = true
            else
                non_dead_ends[room] = true
            end
        end
    )

    local function parent_list(room)
        local parents = {}

        local p = room_parents[room]
        while p and (p.section == room.section) do
            parents[#parents + 1] = p
            p = room_parents[p]
        end

        return parents
    end

    local vertices = Utility.keys_to_array(
      room_graph.vertices, Vector_sort
    )

    local next_key = 2

    local door_pairs = {}

    while (next_key <= (MAX_KEYS + 1)) and (#vertices > 0) do
        local room = Random.remove_from_array(vertices)

        if room_children_count[room] >= MIN_SECTION_CHILDREN then
            local all_parents = parent_list(room)
            if #all_parents > 0 then
                local root = all_parents[#all_parents]

                if (room_children_count[root]
                  - room_children_count[room] - 1)
                  >= MIN_SECTION_CHILDREN then
                    room.section = next_key

                    local edge = Graph.get_edge(
                      room_graph, room, room_parents[room])

                    edge.door = next_key - 1

                    table.insert(door_pairs, {
                        next_key - 1,
                        room,
                        room_parents[room]
                    })

                    for c, _ in pairs(room_children[room]) do
                        c.section = next_key

                        for _, p in ipairs(all_parents) do
                            room_children[p][c] = nil
                            room_children_count[p]
                              = room_children_count[p] - 1
                        end
                    end

                    for _, p in ipairs(all_parents) do
                        room_children[p][room] = nil
                        room_children_count[p]
                          = room_children_count[p] - 1

                        Utility.remove_from_array(vertices, p)
                    end

                    next_key = next_key + 1
                end
            end
        end
    end

    return next_key - 1, dead_ends, non_dead_ends, door_pairs
end

-- Desigate which rooms will contain keys, the exit, treasure, and
-- spawn cubes.  Also removes some dead ends and designates some rooms
-- as corridors.
local function add_room_features(room_graph, dead_ends, non_dead_ends,
  entrance, num_sections)
    local joinable = {}
    local unjoinable = {}
    local non_dead_ends_by_section = {}

    local entrance_table

    local dead_ends = Utility.keys_to_array(dead_ends, Vector_sort)
    local non_dead_ends = Utility.keys_to_array(non_dead_ends,
      Vector_sort)

    local closed_rooms = {}

    for _, room in ipairs(dead_ends) do
        local adjacent = Vector.cardinal_adjacent(room)
        local adj_count = 0
        for _, adj in ipairs(adjacent) do
            local real_adj = coord_to_room_ref(room_graph, adj)
            if real_adj and not Graph.are_adjacent(room_graph, room,
              real_adj) and (real_adj.section == room.section) then
                adj_count = adj_count + 1
            end
        end

        if adj_count > 0 then
            if room == entrance then
                entrance_table = joinable
            else
                Utility.new_tab_insert(
                  joinable, room.section, room
                )
            end
        else
            if room == entrance then
                entrance_table = unjoinable
            else
                Utility.new_tab_insert(
                  unjoinable, room.section, room
                )
            end
        end
    end

    for _, room in ipairs(non_dead_ends) do
        Utility.new_tab_insert(
          non_dead_ends_by_section, room.section, room
        )
    end

    -- Place keys

    local section = 1
    while section < num_sections do
        local room
        local ri
        local room_list

        if unjoinable[section] then
            room_list = unjoinable
        elseif joinable[section] then
            room_list = joinable
        else
            assert(non_dead_ends_by_section[section])
            room_list = non_dead_ends_by_section
        end

        room = Random.remove_from_array(room_list[section])
        if #room_list[section] == 0 then
            room_list[section] = nil
        end

        if room_list == non_dead_ends_by_section then
            Utility.remove_from_array(non_dead_ends, room)
        end

        room.key = section
        closed_rooms[room] = true

        section = section + 1
    end

    -- Place exit

    local exit_room
    while not exit_room or (exit_room == entrance) do
        if unjoinable[num_sections] then
            exit_room
              = Random.remove_from_array(unjoinable[num_sections])
        else
            assert(joinable[num_sections])
            exit_room
              = Random.remove_from_array(joinable[num_sections])
        end
    end

    exit_room.exit = true
    closed_rooms[exit_room] = true

    -- Count entrance as dead end again

    if entrance_table then
        Utility.new_tab_insert(entrance_table, 1, entrance)
    end

    -- Connect dead ends

    local connectable_dead_ends = {}
    local connectable_adj = {}
    for _, rs in pairs(joinable) do
        for _, r in ipairs(rs) do
            local valid_adj = {}

            local current_adj = Graph.get_adjacent(room_graph, r)
            assert(#current_adj == 1)
            current_adj = current_adj[1]

            for _, coord in ipairs(Vector.cardinal_adjacent(r)) do
                local real_adj = coord_to_room_ref(room_graph, coord)
                if real_adj and not closed_rooms[real_adj]
                  and (real_adj ~= current_adj)
                  and (real_adj.section == r.section) then
                    table.insert(valid_adj, real_adj)
                end
            end

            if #valid_adj > 0 then
                table.insert(connectable_dead_ends, r)
                connectable_adj[r] = valid_adj
            else
                Utility.new_tab_insert(unjoinable, r.section, r)
            end
        end
    end

    local num_to_connect = list_percentage(connectable_dead_ends,
      PERCENT_LOOPS)

    while num_to_connect > 0 do
        local room = Random.remove_from_array(
          connectable_dead_ends)

        local current_adj = Graph.get_adjacent(room_graph, room)
        if #current_adj == 1 then
            local adj
              = Random.select_from_array(connectable_adj[room])
            Graph.set_edge(room_graph, room, adj, {})

            if room ~= entrance then
                table.insert(non_dead_ends, room)
                table.insert(non_dead_ends_by_section[room.section],
                  room)
            end

            num_to_connect = num_to_connect - 1

            local adj_i
            _, adj_i = Utility.find_in_array(connectable_dead_ends,
              adj)
            if adj_i then
                table.remove(connectable_dead_ends, adj_i)
                num_to_connect = num_to_connect - 1

                if adj ~= entrance then
                    table.insert(non_dead_ends, adj)
                    table.insert(
                      non_dead_ends_by_section[room.section],
                      adj
                    )
                end
            end
        end
    end

    -- Treasure in remaining dead ends

    for _, room in ipairs(connectable_dead_ends) do
        if room ~= entrance then
            local current_adj = Graph.get_adjacent(room_graph, room)
            if #current_adj == 1 then
                room.treasure = true
            end
        end
    end
    for _, section in pairs(unjoinable) do
        for _, room in ipairs(section) do
            if room ~= entrance then
                room.treasure = true
            end
        end
    end

    -- Spawns

    for i, section in pairs(non_dead_ends_by_section) do
        local room = Random.remove_from_array(section)
        Utility.remove_from_array(
          non_dead_ends,
          room
        )

        while room.entrance do
            room = Random.remove_from_array(section)

            Utility.remove_from_array(
              non_dead_ends,
              room
            )
        end

        if #section == 0 then
            non_dead_ends_by_section[i] = nil
        end

        room.spawn = true
    end

    -- Extra treasure

    local num_extra_treasure = list_percentage(non_dead_ends,
      PERCENT_EXTRA_TREASURE)

    for _ = 1, num_extra_treasure do
        local room = Random.remove_from_array(non_dead_ends)
        room.treasure = true
    end

    -- Corridors

    local num_corridors = list_percentage(non_dead_ends,
      PERCENT_CORRIDOR)
    while num_corridors > 0 do
        local room = Random.remove_from_array(non_dead_ends)
        if room ~= entrance then
            room.corridor = true
            num_corridors = num_corridors - 1
        end
    end
end

-- Selects room pairs and join them into big rooms
local function expand_rooms(room_graph)
    local joins = {}

    local all_room_pairs = Graph.get_all_edges(room_graph)
    local valid_room_pairs = {}

    for _, pair in ipairs(all_room_pairs) do
        local a = pair[1]
        local b = pair[2]
        if (a.section == b.section) and not (a.corridor or b.corridor)
          then
            table.insert(valid_room_pairs, pair)
        end
    end

    local num_joins = list_percentage(valid_room_pairs,
      PERCENT_EXPANDED_ROOM)

    for _ = 1, num_joins do
        local pair = Random.remove_from_array(valid_room_pairs)
        local a = pair[1]
        local b = pair[2]

        local edge = Graph.get_edge(room_graph, a, b)
        edge.join = true
    end
end

-- Select rooms that will have chasms
local function place_chasms(room_graph)
    local vertices = Utility.keys_to_array(
      room_graph.vertices, Vector_sort
    )
    local num_chasms = list_percentage(vertices, PERCENT_CHASM)

    local frontier = {}
    while #frontier < NUM_CHASM_SEEDS do
        local room = Random.remove_from_array(vertices)
        table.insert(frontier, room)
    end

    while (num_chasms > 0) and (#frontier > 0) do
        local room = Random.remove_from_array(frontier)
        if not room.chasm then
            room.chasm = true
            num_chasms = num_chasms - 1

            local adj = Graph.get_adjacent(room_graph, room)
            for _, a in ipairs(adj) do
                table.insert(frontier, a)
            end
        end
    end
end

----------------------------------------------------------------------

-- Place walls and chasms around a room
local function carve_room_walls(map_grid, room_graph, room,
  wall_size, room_half_size)
    local path_exists = {}
    local door_exists = {}
    local join_exists = {}
    local adj_section_exists = {}

    for dir, delta in pairs(Vector.DIRS) do
        local adj = adj_room_ref(room_graph, room, dir)
        if adj then
            if adj.section == room.section then
                adj_section_exists[dir] = adj
            end

            if Graph.are_adjacent(room_graph, room, adj) then
                path_exists[dir] = adj

                local edge = Graph.get_edge(room_graph, room, adj)

                if edge.door then
                    door_exists[dir] = edge.door
                elseif edge.join then
                    join_exists[dir] = adj
                end
            end
        end
    end

    local hall_wall_size = room_half_size - HALL_HALF_SIZE
    local door_wall_size = room_half_size - DOOR_HALF_SIZE

    local min_x = room.x * TOTAL_ROOM_SIZE
    local min_y = room.y * TOTAL_ROOM_SIZE
    local max_x = min_x + TOTAL_ROOM_SIZE - 1
    local max_y = min_y + TOTAL_ROOM_SIZE - 1

    local min_hall_x = min_x + hall_wall_size + wall_size
    local min_hall_y = min_y + hall_wall_size + wall_size
    local max_hall_x = max_x - hall_wall_size - wall_size
    local max_hall_y = max_y - hall_wall_size - wall_size

    local min_door_x = min_x + door_wall_size + wall_size
    local min_door_y = min_y + door_wall_size + wall_size
    local max_door_x = max_x - door_wall_size - wall_size
    local max_door_y = max_y - door_wall_size - wall_size

    local min_room_x = min_x + wall_size
    local min_room_y = min_y + wall_size
    local max_room_x = max_x - wall_size
    local max_room_y = max_y - wall_size

    local at_outer_edge
    local tile

    for d = 0, wall_size - 1 do
        at_outer_edge = d == 0

        for x = min_x + d + 1, max_x - d - 1 do
            -- North

            tile = TILES.wall

            if room.chasm and ((adj_section_exists['north']
              and adj_section_exists['north'].chasm)
              or not at_outer_edge) then
                tile = TILES.chasm
            end

            if path_exists['north'] then
                if (join_exists['north']
                  and (x >= min_room_x) and (x <= max_room_x))
                  or ((x >= min_hall_x) and (x <= max_hall_x)) then
                    tile = TILES.floor
                end

                if door_exists['north'] and at_outer_edge
                  and ((x < min_door_x) or (x > max_door_x)) then
                    tile = TILES.wall
                end
            end

            Grid.set(map_grid, x, min_y + d, tile)

            -- South

            tile = TILES.wall

            if room.chasm and ((adj_section_exists['south']
              and adj_section_exists['south'].chasm)
              or not at_outer_edge) then
                tile = TILES.chasm
            end

            if path_exists['south'] then
                if (join_exists['south']
                  and (x >= min_room_x) and (x <= max_room_x))
                  or ((x >= min_hall_x) and (x <= max_hall_x)) then
                    tile = TILES.floor
                end

                if door_exists['south'] and at_outer_edge
                  and ((x < min_door_x) or (x > max_door_x)) then
                    tile = TILES.wall
                end
            end

            Grid.set(map_grid, x, max_y - d, tile)
        end

        for y = min_y + d + 1, max_y - d - 1 do
            -- East

            tile = TILES.wall

            if room.chasm and ((adj_section_exists['east']
              and adj_section_exists['east'].chasm)
              or not at_outer_edge) then
                tile = TILES.chasm
            end

            if path_exists['east'] then
                if (join_exists['east']
                  and (y >= min_room_y) and (y <= max_room_y))
                  or ((y >= min_hall_y) and (y <= max_hall_y)) then
                    tile = TILES.floor
                end

                if door_exists['east'] and at_outer_edge
                  and ((y < min_door_y) or (y > max_door_y)) then
                    tile = TILES.wall
                end
            end

            Grid.set(map_grid, max_x - d, y, tile)

            -- West

            tile = TILES.wall

            if room.chasm and ((adj_section_exists['west']
              and adj_section_exists['west'].chasm)
              or not at_outer_edge) then
                tile = TILES.chasm
            end

            if path_exists['west'] then
                if (join_exists['west']
                  and (y >= min_room_y) and (y <= max_room_y))
                  or ((y >= min_hall_y) and (y <= max_hall_y)) then
                    tile = TILES.floor
                end

                if door_exists['west'] and at_outer_edge
                  and ((y < min_door_y) or (y > max_door_y)) then
                    tile = TILES.wall
                end
            end

            Grid.set(map_grid, min_x + d, y, tile)
        end

        -- Northeast

        tile = TILES.wall
        if room.chasm
          and ((adj_section_exists['northeast']
            and adj_section_exists['northeast'].chasm
           and adj_section_exists['north']
            and adj_section_exists['north'].chasm
           and adj_section_exists['east']
            and adj_section_exists['east'].chasm)
          or not at_outer_edge) then
            tile = TILES.chasm
        end

        Grid.set(map_grid, max_x - d, min_y + d, tile)

        -- Southeast

        tile = TILES.wall
        if room.chasm
          and ((adj_section_exists['southeast']
            and adj_section_exists['southeast'].chasm
           and adj_section_exists['south']
            and adj_section_exists['south'].chasm
           and adj_section_exists['east']
            and adj_section_exists['east'].chasm)
          or not at_outer_edge) then
            tile = TILES.chasm
        end

        Grid.set(map_grid, max_x - d, max_y - d, tile)

        -- Southwest

        tile = TILES.wall
        if room.chasm
          and ((adj_section_exists['southwest']
            and adj_section_exists['southwest'].chasm
           and adj_section_exists['south']
            and adj_section_exists['south'].chasm
           and adj_section_exists['west']
            and adj_section_exists['west'].chasm)
          or not at_outer_edge) then
            tile = TILES.chasm
        end

        Grid.set(map_grid, min_x + d, max_y - d, tile)

        -- Northwest

        tile = TILES.wall
        if room.chasm
          and ((adj_section_exists['northwest']
            and adj_section_exists['northwest'].chasm
           and adj_section_exists['north']
            and adj_section_exists['north'].chasm
           and adj_section_exists['west']
            and adj_section_exists['west'].chasm)
          or not at_outer_edge) then
            tile = TILES.chasm
        end

        Grid.set(map_grid, min_x + d, min_y + d, tile)
    end
end

-- Track objects such as keys, treasture, the entrance, the exit, and
-- spawns, storing their world coordinates.
local function place_room_feature(map_grid, object_list, room,
  wall_size, room_half_size)
    local result = false

    local min_x = (room.x * TOTAL_ROOM_SIZE) + wall_size
    local min_y = (room.y * TOTAL_ROOM_SIZE) + wall_size
    local max_x = min_x + (room_half_size * 2) - 1
    local max_y = min_y + (room_half_size * 2) - 1

    if room.entrance then
        object_list['entrance'] = Vector.new(
          min_x + room_half_size,
          min_y + room_half_size
        )

        result = true
    elseif room.exit then
        object_list['exit_room'] = Vector.new(
          min_x + room_half_size,
          min_y + room_half_size
        )
        result = true
    elseif room.key then
        Utility.set_double_key(
            object_list,
            'keys', room.key,
            Vector.new(
              min_x + room_half_size,
              min_y + room_half_size
            )
        )

        result = true
    elseif room.spawn then
        Utility.set_double_key(
            object_list,
            'spawn', room.section,
            Vector.new(
              min_x + room_half_size,
              min_y + room_half_size
            )
        )

        result = true
    elseif room.treasure then
        Utility.new_tab_insert(
            object_list, 'treasure',
            Vector.new(
              min_x + room_half_size,
              min_y + room_half_size
            )
        )

        result = true
    end

    return result
end

-- Place the interior floor and walls of a room.  Place monsters.
local function carve_room(map_grid, object_list, room, wall_size,
  room_half_size, carve_pattern)
    local monsters = {}

    local min_x = (room.x * TOTAL_ROOM_SIZE) + wall_size
    local min_y = (room.y * TOTAL_ROOM_SIZE) + wall_size
    local max_x = min_x + (room_half_size * 2) - 1
    local max_y = min_y + (room_half_size * 2) - 1

    for x = min_x, max_x do
        for y = min_y, max_y do
            Grid.set(map_grid, x, y, TILES.floor)
        end
    end

    local feature_placed = place_room_feature(map_grid, object_list,
      room, wall_size, room_half_size)

    if carve_pattern then
        local reflect
          = Random.select_from_array(ROOM_PATTERN_REFLECTIONS)

        local pattern_max_x
        local pattern_max_y

        if reflect.h then
            pattern_max_y = room_half_size - 1
            if feature_placed then
                pattern_max_y = pattern_max_y - 1
            end
        else
            pattern_max_y = (room_half_size * 2) - 2
        end

        if reflect.v then
            pattern_max_x = room_half_size - 1
            if feature_placed then
                pattern_max_x = pattern_max_x - 1
            end
        else
            pattern_max_x = (room_half_size * 2) - 2
        end

        local possible_tiles = {}

        for x = 1, pattern_max_x do
            for y = 1, pattern_max_y do
                table.insert(possible_tiles, Vector.new(x, y))
            end
        end

        local num_new_walls = list_percentage(possible_tiles,
          PERCENT_INTERIOR_WALL)

        for _ = 1, num_new_walls do
            local tile = Random.remove_from_array(possible_tiles)

            Grid.set(map_grid, min_x + tile.x, min_y + tile.y,
              TILES.inner_wall)

            if reflect.h then
                Grid.set(map_grid, min_x + tile.x, max_y - tile.y,
                  TILES.inner_wall)
            end
            if reflect.v then
                Grid.set(map_grid, max_x - tile.x, min_y + tile.y,
                  TILES.inner_wall)
            end
            if reflect.h and reflect.v then
                Grid.set(map_grid, max_x - tile.x, max_y - tile.y,
                  TILES.inner_wall)
            end
        end

        -- Remove inaccessible floors

        local accessible = {}

        if reflect.h and not reflect.v then
            pattern_max_x = pattern_max_x + 1
        end
        if reflect.v and not reflect.h then
            pattern_max_y = pattern_max_y + 1
        end

        Grid.flood_fill(
            map_grid,
            min_x, min_y,
            min_x, min_y,
            min_x + pattern_max_x, min_y + pattern_max_y,
            function (x, y)
                Utility.set_double_key(accessible, x, y, true)
            end,
            function (x, y)
                return Grid.get(map_grid, x, y) ~= TILES.inner_wall
            end
        )

        if reflect.h and not reflect.v then
            pattern_max_x = pattern_max_x - 1
        end
        if reflect.v and not reflect.h then
            pattern_max_y = pattern_max_y - 1
        end

        for x = 1, pattern_max_x do
            for y = 1, pattern_max_y do
                local test_x = min_x + x
                local test_y = min_y + y
                if Grid.get(map_grid, test_x, test_y)
                   ~= TILES.inner_wall
                  and not (accessible[test_x]
                  and accessible[test_x][test_y]) then
                    Utility.remove_from_array(
                      possible_tiles, Vector.new(x, y),
                      Vector.equals
                    )

                    Grid.set(map_grid, min_x + x, min_y + y,
                      TILES.inner_wall)

                    if reflect.h then
                        Grid.set(map_grid, min_x + x, max_y - y,
                          TILES.inner_wall)
                    end
                    if reflect.v then
                        Grid.set(map_grid, max_x - x, min_y + y,
                          TILES.inner_wall)
                    end
                    if reflect.h and reflect.v then
                        Grid.set(map_grid, max_x - x, max_y - y,
                          TILES.inner_wall)
                    end
                end
            end
        end

        -- Place monsters

        if not room.entrance then
            local monster_type
              = Random.select_from_array(MONSTER_TYPES)
            local num_monsters = list_percentage(possible_tiles,
              PERCENT_MONSTERS)
            if num_monsters == 0 then
                num_monsters = 1
            end

            for _ = 1, num_monsters do
                local tile = Random.remove_from_array(possible_tiles)

                assert(Grid.get(
                  map_grid, min_x + tile.x, min_y + tile.y)
                    ~= TILES.inner_wall
                )

                table.insert(monsters, {
                    name = monster_type,
                    x = min_x + tile.x, y = min_y + tile.y,
                    section = room.section
                })

                if reflect.h then
                    assert(Grid.get(
                      map_grid, min_x + tile.x, max_y - tile.y)
                        ~= TILES.inner_wall
                    )

                    table.insert(monsters, {
                        name = monster_type,
                        x = min_x + tile.x, y = max_y - tile.y,
                        section = room.section
                    })
                end
                if reflect.v then
                    assert(Grid.get(
                      map_grid, max_x - tile.x, min_y + tile.y)
                        ~= TILES.inner_wall
                    )

                    table.insert(monsters, {
                        name = monster_type,
                        x = max_x - tile.x, y = min_y + tile.y,
                        section = room.section
                    })
                end
                if reflect.h and reflect.v then
                    assert(Grid.get(
                      map_grid, max_x - tile.x, max_y - tile.y)
                        ~= TILES.inner_wall
                    )

                    table.insert(monsters, {
                        name = monster_type,
                        x = max_x - tile.x, y = max_y - tile.y,
                        section = room.section
                    })
                end
            end
        end
    end

    return monsters
end

-- Generates the tiles for every room and the world coordinates of all
-- objects and monsters
local function carve_map_grid(room_graph, dungeon_bounds)
    local width = TOTAL_ROOM_SIZE * dungeon_bounds.x
    local height = TOTAL_ROOM_SIZE * dungeon_bounds.y

    local map_grid = Grid.new(width, height)
    local object_list = {}
    local monsters = {}

    for room, _ in pairs(room_graph.vertices) do
        local wall_size
        local room_half_size
        local door_half_size
        local carve_pattern

        if room.corridor then
            wall_size = WALL_SIZE + ROOM_HALF_SIZE + CHASM_PADDING
              - HALL_HALF_SIZE
            room_half_size = HALL_HALF_SIZE
            carve_pattern = false
        else
            if room.chasm then
                wall_size = WALL_SIZE + CHASM_PADDING
                room_half_size = ROOM_HALF_SIZE
            else
                wall_size = WALL_SIZE
                room_half_size = FULL_ROOM_HALF_SIZE
            end

            carve_pattern = true
        end

        carve_room_walls(map_grid, room_graph, room,
          wall_size, room_half_size)
        local room_monsters = carve_room(map_grid, object_list, room,
          wall_size, room_half_size, carve_pattern)

        for _, m in ipairs(room_monsters) do
            table.insert(monsters, m)
        end
    end

    return map_grid, object_list, monsters
end

----------------------------------------------------------------------

-- Get world positions and orientations of all the doors
local function get_doors(door_pairs, object_list)
    local door_coords = {}

    for _, door in ipairs(door_pairs) do
        local colour = door[1]
        local door_a = door[2]
        local door_b = door[3]

        local dir_name = Vector.get_dir(door_a, door_b)

        local is_horizontal
          = (dir_name == 'north') or (dir_name == 'south')
        assert(is_horizontal
          or ((dir_name == 'east') or (dir_name == 'west')))

        local start_door = door_a
        if (dir_name == 'west') or (dir_name == 'south') then
            start_door = door_b
            dir_name = Vector.get_dir(door_b, door_a)
        end

        local dir = Vector.CARDINAL_DIRS[dir_name]

        local x = (TOTAL_ROOM_SIZE * start_door.x) +
          FULL_ROOM_HALF_SIZE + WALL_SIZE
        local y = (TOTAL_ROOM_SIZE * start_door.y) +
          FULL_ROOM_HALF_SIZE + WALL_SIZE

        if not is_horizontal then
            x = x + ((FULL_ROOM_HALF_SIZE + WALL_SIZE) * dir.x)
        else
            y = y + ((FULL_ROOM_HALF_SIZE + WALL_SIZE) * dir.y)
        end

        door_coords[colour] = {
            is_horizontal = is_horizontal,
            pos = Vector.new(x, y)
        }
    end

    object_list['doors'] = door_coords
end

----------------------------------------------------------------------

-- Place a wall tile on the map's top layer and, if necessary, on the
-- maps main layer
local function place_wall_tile(map, map_grid, x, y,
  front_name, top_name)
    local south_tile = Grid.get(map_grid, x, y + 1)
    if south_tile and ((south_tile == TILES.wall)
      or (south_tile == TILES.inner_wall)) then
        Map.set_tile(map, x, y, top_name, LAYER_WALL_TOP)
    else
        Map.set_tile(map, x, y, front_name, LAYER_WALL_FRONT)
        Map.set_tile(map, x, y, top_name, LAYER_WALL_TOP)
    end
end

-- Place a wall front or floor edge tile over a chasm
local function place_cliff_tile(map, map_grid, x, y)
    local north_tile = Grid.get(map_grid, x, y - 1)
    if north_tile and not (north_tile == TILES.chasm) then
        local new_tile
        if north_tile == TILES.wall then
            new_tile = 'wall_front'
        else
            new_tile = 'floor_edge'
        end
        Map.set_tile(map, x, y, new_tile, LAYER_CLIFF)
    end
end

-- Translate the basic map grid into an actual tile map ready
local function set_map_tiles(map, map_grid)
    local tile
    local north_tile
    local south_tile

    for x = 0, map_grid.width - 1 do
        for y = 0, map_grid.height do -- Extra height for south
                                      -- exterior walls
            tile = Grid.get(map_grid, x, y)
            if tile and not (tile == TILES.chasm) then
                if tile == TILES.floor then
                    Map.set_tile(map, x, y, 'floor', LAYER_FLOOR)
                elseif tile == TILES.wall then
                    place_wall_tile(map, map_grid, x, y,
                      'wall_front', 'wall_top')
                elseif tile == TILES.inner_wall then
                    place_wall_tile(map, map_grid, x, y,
                      'inner_wall_front', 'inner_wall_top')
                else
                    assert(false, 'Unexpected tile value: ' .. tile)
                end
            else
                place_cliff_tile(map, map_grid, x, y)
            end
        end
    end
end

-- Place the four tiles in a room that form a platform where special
-- features (keys, mystery boxes, etc.) are placed
local function set_map_feature_square(map, pos)
    Map.set_tile(map, pos.x - 1, pos.y - 1, 'special_floor_nw',
      LAYER_FLOOR)
    Map.set_tile(map, pos.x    , pos.y - 1, 'special_floor_ne',
      LAYER_FLOOR)
    Map.set_tile(map, pos.x    , pos.y    , 'special_floor_se',
      LAYER_FLOOR)
    Map.set_tile(map, pos.x - 1, pos.y    , 'special_floor_sw',
      LAYER_FLOOR)
end

-- For all rooms with special features and objects in them, place the
-- feature platform tiles inside them
local function set_map_feature_tiles(map, object_list)
    set_map_feature_square(map, object_list['entrance'])
    set_map_feature_square(map, object_list['exit_room'])

    for _, pos in pairs(object_list['keys']) do
        set_map_feature_square(map, pos)
    end
    for _, pos in pairs(object_list['spawn']) do
        set_map_feature_square(map, pos)
    end
    for _, pos in pairs(object_list['treasure']) do
        set_map_feature_square(map, pos)
    end
end

----------------------------------------------------------------------

-- Generate a map.  Returns the following:
--     The map itself
--     A simplified version of the map for AI pathfinding
--     A logical graph of the map's rooms
--     A table of all special objects and features, containing their
--         locations
--     A table of all monsters and their locations
local function generate()
    local room_graph, dungeon_bounds = room_graph()
    local entrance = set_entrance(room_graph, dungeon_bounds)

    local num_sections, dead_ends, non_dead_ends, door_pairs
      = create_sections(room_graph, entrance)

    add_room_features(room_graph, dead_ends,
      non_dead_ends, entrance, num_sections)
    expand_rooms(room_graph)
    place_chasms(room_graph)

    local map_grid
    local object_list
    local monsters
    map_grid, object_list, monsters
      = carve_map_grid(room_graph, dungeon_bounds)

    get_doors(door_pairs, object_list)

    -- Extra height to allow cliff tiles along the south exterior
    -- walls
    local map = Map.new(map_grid.width, map_grid.height + 1)

    Map.add_tileset(map, 'dungeon')
    Map.set_object_layer(map, LAYER_WALL_FRONT)

    Map.set_layer_offset(map, LAYER_WALL_FRONT, OFFSET_WALL_FRONT)
    Map.set_layer_offset(map, LAYER_WALL_TOP, OFFSET_WALL_TOP)

    set_map_tiles(map, map_grid)
    set_map_feature_tiles(map, object_list)

    Map.generate_walls(
        map,
        function(x, y)
            return (Grid.get(map_grid, x, y) == TILES.wall)
              or (Grid.get(map_grid, x, y) == TILES.inner_wall)
        end
    )
    Map.generate_walls(
        map,
        function(x, y)
            return Grid.get(map_grid, x, y) == TILES.chasm
        end,
        Constants.PHYS_CAT_CHASM
    )

    return map, map_grid, room_graph, object_list, monsters
end
Dungeon.generate = generate

----------------------------------------------------------------------

return Dungeon