--
-- Copyright 2012 Aaron MacDonald
--
-- Licensed under the Apache License, Version 2.0 (the "License");
-- you may not use this file except in compliance with the License.
-- You may obtain a copy of the License at
--
--     http://www.apache.org/licenses/LICENSE-2.0
--
-- Unless required by applicable law or agreed to in writing, software
-- distributed under the License is distributed on an "AS IS" BASIS,
-- WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
-- implied.
-- See the License for the specific language governing permissions and
-- limitations under the License.
--

local Game = {}

local Settings = require 'settings'

local Constants = require 'constants'
local Vector = require 'vector'
local Utility = require 'utility'
local Random = require 'random'

local ScreenManager = require 'screen_manager'
local MouseManager = require 'mouse_manager'
local SoundManager = require 'sound_manager'

local Map = require 'map'
local Dungeon = require 'dungeon'
local Graph = require 'graph'
local Grid = require 'grid'

local Parallax = require 'parallax'
local ColourFlux = require 'colour_flux'

local Menu = require 'menu'

local Entrance = require 'entrance'
local Door = require 'door'
local Key = require 'key'
local MysteryBox = require 'mystery_box'
local Spawn = require 'spawn'

local Actor = require 'actor'
local Bullet = require 'bullet'
local AI = require 'ai'

local Ghost = require 'ghost'
local Particles = require 'particles'
local AimGuide = require 'aim_guide'

local PersistantGameData = require 'persistant_data'

local ActorData = require 'data/actors'

----------------------------------------------------------------------

local GUI_MARGIN = 16
local GUI_INNER_MARGIN = 8

local GUI_ALPHA = 192

local KEY_AREA_WIDTH = (Key.FULL_WIDTH * Constants.MAX_DOORS)
  + (GUI_INNER_MARGIN * (Constants.MAX_DOORS - 1)) + (GUI_MARGIN * 2)

local HEALTHBAR_HEIGHT = 32
local HEALTHBAR_Y_ADD = (Key.FULL_HEIGHT - HEALTHBAR_HEIGHT) / 2

local HEALTHBAR_FILL_COLOUR = {0, 255, 128, GUI_ALPHA}
local HEALTHBAR_FILL_INVINCIBLE_COLOUR = {0, 128, 255, GUI_ALPHA}
local HEALTHBAR_BACK_COLOUR = {0, 0, 128, GUI_ALPHA}

local MINIMAP_ROOM_SIZE = 16
local MINIMAP_ROOM_MARGIN = 4
local MINIMAP_PLAYER_POS_RADIUS = 4
local MINIMAP_SPAWN_RADIUS = 6
local MINIMAP_EXIT_RADIUS = 6
local MINIMAP_DOOR_SIZE = 6

local MINIMAP_ROOM_SPACE = MINIMAP_ROOM_SIZE
  + (MINIMAP_ROOM_MARGIN * 2)

local MINIMAP_ROOM_MIDDLE
  = (MINIMAP_ROOM_SIZE / 2) + MINIMAP_ROOM_MARGIN

local MINIMAP_DOOR_BORDER
  = (MINIMAP_ROOM_SPACE - MINIMAP_DOOR_SIZE) / 2

local AI_DELAY = 30

local GAME_OVER_DELAY = 180

local TRIPLE_SHOT_ANGLE = 9 * (math.pi / 180)
local NUM_TRIPLE_SHOTS = 64

local NUM_EXPLODING_SHOTS = 64

local INVINCIBILITY_TIME = 900

local OVERLAY_TIME = 60

local FREEZE_OVERLAY_COLOUR = {192, 192, 255}
local SMARTBOMB_OVERLAY_COLOUR = {255, 64, 0}
local SPAWN_DESTROY_OVERLAY_COLOUR = {64, 255, 128}

local DOOR_OPEN_SOUND = 'door_open'
local KEY_PICKUP_SOUND = 'pickup_key'
local BOX_PICKUP_SOUND = 'pickup_box'

local POWER_HEAL_SOUND = 'power_heal'
local POWER_EXPLODING_SHOTS_SOUND = 'power_exploding_shots'
local POWER_TRIPLE_SHOTS_SOUND = 'power_triple_shots'
local POWER_FLIGHT_SOUND = 'power_flight'
local POWER_FREEZE_SOUND = 'power_freeze'
local POWER_SMARTBOMB_SOUND = 'power_smartbomb'

local SPAWN_DESTROY_SOUND = 'spawn_destroy'

----------------------------------------------------------------------

local GUI_TEXT_SIZE = 32
local gui_font
local gui_font_height

local ICON_TRIPLE_SHOT_PATH = 'icon_triple_shot'
local icon_triple_shot

local ICON_EXPLODING_SHOT_PATH = 'icon_exploding_shot'
local icon_exploding_shot

local MINIMAP_ROOM_COLOUR = {128, 128, 255, GUI_ALPHA}
local MINIMAP_ROOM_HIGHLIGHT_COLOUR = {192, 192, 255, GUI_ALPHA}
local CURRENT_ROOM_COLOUR_SPEED = 0.05

local current_room_flux

local overlay_colour
local overlay_time

----------------------------------------------------------------------

local INNER_AMBIENCE_SOUND = 'inner_ambience'
local CHASM_AMBIENCE_SOUND = 'chasm_ambience'
local AMBIENCE_FADE_TIME = 60

local inner_ambience_sound
local chasm_ambience_sound

local current_ambience
local old_ambience
local ambience_fade_time

----------------------------------------------------------------------

local current_map
local map_grid
local map_graph

local abyss

local doors
local door_tiles

local entrance
local exit

local keys
local collected_keys
local unlocked_regions

local mystery_boxes
local spawns

local player
local player_pos

local player_flight_particles

local monsters

local ghosts

local ai_delay

local bullets
local bullet_explosions

local particles

local exit_found
local game_over_timer

local paused
local quit_game

----------------------------------------------------------------------

local PAUSE_CHOICES = {
    {
        text = 'Resume',
        func = function()
            paused = false
            SoundManager.resume()

			current_ambience:resume()
			if old_ambience then
				old_ambience:resume()
			end
        end
    },
    {
        text = 'Quit to Title',
        func = function()
            quit_game = true
        end
    }
}
local pause_menu

local MENU_CLICKED_DELAY = 10
local menu_just_clicked

----------------------------------------------------------------------

local processed_colliders
local removed_colliders

local fixtures_to_remove
local bodies_to_remove

----------------------------------------------------------------------

local COLLISION_DISPATCHER = {}
local function double_dispatch(a, b, func)
    Utility.set_double_key(COLLISION_DISPATCHER, a, b, func)
    if a ~= b then
        Utility.set_double_key(
            COLLISION_DISPATCHER, b, a,
            function(x, y, coll) func(y, x, coll) end
        )
    end
end

----------------------------------------------------------------------

local function remove_bullet(b, normal, target_actor)
    table.insert(fixtures_to_remove, b.phys_data.fixture)
    table.insert(bodies_to_remove, b.phys_data.body)

    Utility.remove_from_array(bullets, b)
    removed_colliders[b] = true

    local pos = Bullet.pos(b)
    local explosion = Particles.create_explosion(
      pos.x, pos.y, 64, 64, 30)
    explosion:setColors(
        unpack(ActorData[b.actor.name].bullet_explosion_colours)
    )

    table.insert(particles, explosion)

    if b.exploding_shot then
        local actors_to_ignore = {}
        if target_actor then
            actors_to_ignore[target_actor] = true
        end

        local p = pos
        if normal then
            p = Vector.add(p, normal)
        end
        table.insert(bullet_explosions, {
            pos = p, actor = b.actor,
            actors_to_ignore = actors_to_ignore
        })
    end
end

local function hit_actor(a, damage)
    local damage_done = Actor.hit(a, damage)
    if not Actor.is_alive(a) then
        for _, f in ipairs(a.fixtures) do
            table.insert(fixtures_to_remove, f)
        end
        table.insert(bodies_to_remove, a.body)

        removed_colliders[a] = true

        -- Create explosion

        local pos = Actor.pos(a)
        local start_colour = ActorData[a.name].start_explosion_colour
        local end_colour = ActorData[a.name].end_explosion_colour
        local explosion = Particles.create_explosion(
          pos.x, pos.y, 96, 96, 45)
        explosion:setColors(
            unpack(ActorData[a.name].explosion_colours)
        )

        table.insert(particles, explosion)

        if a == player then
            table.insert(ghosts, Ghost.new(Actor.pos(a), a.name))
            player_flight_particles = nil
            game_over_timer = GAME_OVER_DELAY
        else
            table.insert(ghosts, Ghost.new(
                Actor.pos(a), a.name, spawns[a.ai_data.section]
            ))
            Utility.remove_from_array(monsters, a)
        end
    end
    return damage_done
end

----------------------------------------------------------------------

double_dispatch('actor', 'key', function(actor, key)
    if actor == player then
        table.insert(fixtures_to_remove, key.phys_data.fixture)
        table.insert(bodies_to_remove, key.phys_data.body)
        removed_colliders[key] = true

        keys[key.index] = nil
        table.insert(collected_keys, key.index)

        local pos = Actor.pos(actor)
        SoundManager.new_sound(KEY_PICKUP_SOUND, pos.x, pos.y)
    end
end)

double_dispatch('actor', 'mystery_box', function(actor, box)
    if actor == player then
        table.insert(fixtures_to_remove, box.phys_data.fixture)
        table.insert(bodies_to_remove, box.phys_data.body)
        removed_colliders[box] = true

        Utility.remove_from_array(mystery_boxes, box)
        PersistantGameData.add_mystery_box()

        local pos = Actor.pos(actor)
        SoundManager.new_sound(BOX_PICKUP_SOUND, pos.x, pos.y)
    end
end)

double_dispatch('actor', 'door', function(actor, door)
    if (actor == player) and not keys[door.index] then
        unlocked_regions[door.index + 1] = true

        local tiles = Door.covered_tiles(door)
        for _, t in ipairs(tiles) do
            Grid.set(door_tiles, t.x, t.y, false)
        end

        door.open = true

        table.insert(fixtures_to_remove, door.phys_data.fixture)
        table.insert(bodies_to_remove, door.phys_data.body)
        removed_colliders[door] = true

        SoundManager.new_sound(DOOR_OPEN_SOUND,
          door.pos.x, door.pos.y)
    end
end)

double_dispatch('actor', 'exit', function(actor, key)
    if actor == player then
        exit_found = true
    end
end)

double_dispatch('bullet', 'door', function(bullet, door, coll)
    Bullet.hit_wall(bullet)
    local normal = Vector.new(coll:getNormal())
    Vector.scale_inline(normal, Bullet.RADIUS / Constants.TILE_SIZE)
    remove_bullet(bullet, normal)
end)

double_dispatch('bullet', 'wall', function(bullet, wall, coll)
    Bullet.hit_wall(bullet)
    local normal = Vector.new(coll:getNormal())
    Vector.scale_inline(normal, Bullet.RADIUS / Constants.TILE_SIZE)
    remove_bullet(bullet, normal)
end)

double_dispatch('hitbox', 'hitbox', function(actor_a, actor_b)
    local plr
    local attacker

    if actor_a == player then
        plr = actor_a
        attacker = actor_b
    elseif actor_b == player then
        plr = actor_b
        attacker = actor_a
    end

    if plr and (ai_delay == 0) and (attacker.freeze_time == 0) then
        local damage_done = hit_actor(
            plr, ActorData[attacker.name].damage
        )
        if damage_done then
            local p = Actor.pos(attacker)
            SoundManager.new_sound(
                ActorData[attacker.name].attack_sound,
                p.x, p.y
            )
      end
    end
end)

double_dispatch('bullet', 'hitbox', function(bullet, actor)
    local player_bullet = bullet.actor == player
    local player_target = actor == player

    local player_hit = not player_bullet and player_target
    local monster_hit = player_bullet and not player_target

    local ignore_hit = bullet.actors_to_ignore
      and  bullet.actors_to_ignore[actor]

    if not ignore_hit and (player_hit or monster_hit) then
        local normal = Vector.normal(
            Vector.new(bullet.phys_data.body:getLinearVelocity())
        )
        Vector.scale_inline(normal, Bullet.RADIUS)
        normal.x = -normal.x
        normal.y = -normal.y

        local damage_done = hit_actor(actor, Bullet.damage(bullet))
        Bullet.hit_actor(bullet, damage_done)
        remove_bullet(bullet, normal, actor)
    end
end)

local function collision_start(a, b, coll)
    local ad = a:getUserData()
    local bd = b:getUserData()

    assert(ad.type)
    assert(ad.data)
    assert(bd.type)
    assert(bd.data)

    if not removed_colliders[ad.data]
       and not removed_colliders[bd.data]
      and not (processed_colliders[ad.data]
       and processed_colliders[ad.data][bd.data])
      and not (processed_colliders[bd.data]
       and processed_colliders[bd.data][ad.data])
      and COLLISION_DISPATCHER[ad.type]
      and COLLISION_DISPATCHER[ad.type][bd.type] then
        Utility.set_double_key(processed_colliders, ad.data, bd.data,
          true)
        COLLISION_DISPATCHER[ad.type][bd.type](ad.data, bd.data, coll)
    end
end

----------------------------------------------------------------------

local function load()
    abyss = Parallax.new({
        {image_name = 'circuits_small', speed = 0.167},
        {image_name = 'circuits_med', speed = 0.333},
        {image_name = 'circuits_large', speed = 0.5}
    })

    current_room_flux = ColourFlux.new(
        {MINIMAP_ROOM_COLOUR, MINIMAP_ROOM_HIGHLIGHT_COLOUR},
        CURRENT_ROOM_COLOUR_SPEED
    )

    processed_colliders = {}
    removed_colliders = {}
    fixtures_to_remove = {}
    bodies_to_remove = {}

    gui_font = love.graphics.newFont(Constants.DEFAULT_FONT,
      GUI_TEXT_SIZE)
    gui_font:setLineHeight(1.5)
    gui_font_height = gui_font:getHeight()

    icon_triple_shot = love.graphics.newImage(
      Utility.img_path(ICON_TRIPLE_SHOT_PATH))
    icon_exploding_shot = love.graphics.newImage(
      Utility.img_path(ICON_EXPLODING_SHOT_PATH))

    pause_menu = Menu.new(0, 0, gui_font, PAUSE_CHOICES)



    inner_ambience_sound = love.audio.newSource(Constants.SOUND_PATH
      .. '/' .. INNER_AMBIENCE_SOUND .. '.'
      .. Constants.DEFAULT_SOUND_TYPE, 'stream')
    chasm_ambience_sound = love.audio.newSource(Constants.SOUND_PATH
      .. '/' .. CHASM_AMBIENCE_SOUND .. '.'
      .. Constants.DEFAULT_SOUND_TYPE, 'stream')

    inner_ambience_sound:setLooping(true)
    chasm_ambience_sound:setLooping(true)

    AimGuide.load()
    PersistantGameData.load()
end
Game.load = load

local function init()
    collected_keys = {}
    unlocked_regions = {}
    unlocked_regions[1] = true

    local object_list
    local monster_data

    current_map, map_grid, map_graph, object_list, monster_data
      = Dungeon.generate()

    Map.set_collision_callbacks(current_map, collision_start)

    local exit_pos = Vector.scale(object_list['exit_room'],
      Constants.TILE_SIZE)
    exit = Entrance.new_exit(current_map.phys_world, exit_pos)

    doors = {}
    door_tiles = Grid.new(map_grid.width, map_grid.height)

    local door_data = object_list['doors']
    for colour, door in pairs(door_data) do
        local d = Door.new(
          current_map.phys_world,
          colour,
          door.is_horizontal,
          Vector.scale(door.pos, Constants.TILE_SIZE)
        )
        table.insert(doors, d)

        local tiles = Door.covered_tiles(d)
        for _, t in ipairs(tiles) do
            Grid.set(door_tiles, t.x, t.y, true)
        end
    end

    keys = {}

    local key_data = object_list['keys']
    for colour, k in pairs(key_data) do
        keys[colour] = Key.new(
          current_map.phys_world,
          colour,
          Vector.scale(k, Constants.TILE_SIZE)
        )
    end

    mystery_boxes = {}

    local box_data = object_list['treasure']
    for _, b in ipairs(box_data) do
        table.insert(mystery_boxes, MysteryBox.new(
          current_map.phys_world, Vector.scale(b, Constants.TILE_SIZE)
        ))
    end

    spawns = {}

    local spawn_data = object_list['spawn']
    for section, s in pairs(spawn_data) do
        spawns[section] = Spawn.new(
          current_map, section, Vector.scale(s, Constants.TILE_SIZE))
    end

    local entrance_pos = Vector.scale(object_list['entrance'],
      Constants.TILE_SIZE)
    entrance = Entrance.new_entrance(entrance_pos)

    player = Actor.new('player', current_map,
      entrance_pos.x, entrance_pos.y)
    player.health = PersistantGameData.current_player_health()
    player.triple_shots
      = PersistantGameData.current_num_triple_shots()
    player.exploding_shots
      = PersistantGameData.current_num_exploding_shots()
    Actor.make_invincible(player,
      PersistantGameData.current_invincibility_time(),
      true
    )

    AimGuide.set_player_shoot_y(player.shoot_y)

    player_pos = Actor.pos(player)
    SoundManager.set_listener(player_pos.x, player_pos.y)

    monsters = {}

    for _, m in ipairs(monster_data) do
        local mon = Actor.new(
            m.name,
            current_map,
            (m.x * Constants.TILE_SIZE) + Constants.TILE_HALF_SIZE,
            (m.y * Constants.TILE_SIZE) + Constants.TILE_HALF_SIZE
        )
        mon.ai_data = {
            section = m.section,
            wandering = true
        }

        assert(spawns[mon.ai_data.section], mon.ai_data.section)
        table.insert(monsters, mon)
    end

    ghosts = {}

    ai_delay = AI_DELAY

    bullets = {}
    bullet_explosions = {}
    particles = {}

    exit_found = false

    overlay_colour = nil
    overlay_time = 0

    paused = false
    quit_game = false
    menu_just_clicked = 0

    PersistantGameData.new_level()
end
Game.init = init

----------------------------------------------------------------------

local function end_game()
    current_map = nil
    map_grid = nil
    map_graph = nil

    doors = nil
    door_tiles = nil

    entrance = nil
    exit = nil

    keys = nil
    collected_keys = nil
    unlocked_regions = nil

    mystery_boxes = nil
    spawns = nil

    player = nil
    player_pos = nil

    player_flight_particles = nil

    monsters = nil

    ghosts = nil

    ai_delay = nil

    bullets = nil
    bullet_explosions = nil
    particles = nil

    exit_found = nil
    game_over_timer = nil

    paused = false

    if current_ambience then
        current_ambience:stop()
        current_ambience = nil
    end
    if old_ambience then
        old_ambience:stop()
        old_ambience = nil
    end
    ambience_fade_time = nil

    Parallax.reset(abyss)
    ColourFlux.reset(current_room_flux)

    processed_colliders = {}
    fixtures_to_remove = {}
    removed_colliders = {}
    bodies_to_remove = {}

    overlay_colour = nil
    overlay_time = 0

    AI.reset_globals()
    SoundManager.clear()
    SoundManager.reset_listener()
end

----------------------------------------------------------------------

local function set_overlay(colour)
    overlay_colour = colour
    overlay_time = OVERLAY_TIME
end

-- Healing

local function box_can_heal()
    return Actor.health_percentage(player) < 1
end

local function box_heal()
    local max = ActorData['player'].health
    local missing = max - player.health
    Actor.heal(player, missing)

    local p = Actor.pos(player)
    SoundManager.new_sound(POWER_HEAL_SOUND, p.x, p.y)
end

-- Triple shot

local function box_can_triple_shot()
    return player.triple_shots == 0
end

local function box_triple_shot()
    player.triple_shots = NUM_TRIPLE_SHOTS

    local p = Actor.pos(player)
    SoundManager.new_sound(POWER_TRIPLE_SHOTS_SOUND, p.x, p.y)
end

-- Exploding shot

local function box_can_exploding_shot()
    return player.exploding_shots == 0
end

local function box_exploding_shot()
    player.exploding_shots = NUM_EXPLODING_SHOTS

    local p = Actor.pos(player)
    SoundManager.new_sound(POWER_EXPLODING_SHOTS_SOUND, p.x, p.y)
end

-- Flight

local function box_can_fly()
    return not player.flies
end

local function box_fly()
    Actor.make_flyable(player)

    player_flight_particles = Particles.create_fountain(
        0, 0, 500, 100, 64, 50, 100
    )
    player_flight_particles:setColors(
        128, 255, 255, 255,
        128, 255, 128, 128
    )
    player_flight_particles:start()

    local p = Actor.pos(player)
    SoundManager.new_sound(POWER_FLIGHT_SOUND, p.x, p.y)
end

-- Invincibility

local function box_can_invincibility()
    return player.invincibility_time == 0
end

local function box_invincibility()
    Actor.make_invincible(player, INVINCIBILITY_TIME)
end

-- Monster freezing

local function box_can_freeze()
    local result = true
    for _, m in ipairs(monsters) do
        if m.freeze_time > 0 then
            result = false
            break
        end
    end
    return result
end

local function box_freeze()
    for _, m in ipairs(monsters) do
        Actor.freeze(m)
    end

    set_overlay(FREEZE_OVERLAY_COLOUR)
    SoundManager.new_sound(POWER_FREEZE_SOUND)
end

-- Monster destroying

local function box_can_smartbomb()
    return #monsters > 0
end

local function box_smartbomb()
    while #monsters > 0 do
        local m = monsters[1]
        hit_actor(m, m.health)
    end

    set_overlay(SMARTBOMB_OVERLAY_COLOUR)
    SoundManager.new_sound(POWER_SMARTBOMB_SOUND)
end

-- Spawn cube destroying

local function box_can_destroy_spawn()
    local result = false

    for k, _ in pairs(spawns) do
        if unlocked_regions[k] then
            result = true
            break
        end
    end

    return result
end

local function box_destroy_spawn()
    local keys = {}
    for k, _ in pairs(spawns) do
        if unlocked_regions[k] then
            table.insert(keys, k)
        end
    end

    local target_index = Random.select_from_array(keys)
    local target_spawn = spawns[target_index]

    spawns[target_index] = nil

    Spawn.destroy(target_spawn)

    local pos = target_spawn.pos

    -- The explosion

    set_overlay(SPAWN_DESTROY_OVERLAY_COLOUR)

    local radius = map_grid.width
    if map_grid.height > map_grid.width then
        radius = map_grid.height
    end
    radius = radius * Constants.TILE_SIZE

    local explosion = Particles.create_explosion(
      pos.x, pos.y, 7500, radius, 120)
    explosion:setColors(
          0, 255, 128, 255,
        255, 255, 255,   0
    )
    explosion:setSizes(2)

    table.insert(particles, explosion)

    SoundManager.new_sound(SPAWN_DESTROY_SOUND, pos.x, pos.y, radius)

    -- The lost ghosts

    for _ = 1, #target_spawn.queue do
        local range = 64 * math.random()
        local angle = math.pi * 2 * math.random()

        local x = (range * math.cos(angle)) + pos.x
        local y = (range * math.sin(angle)) + pos.y

        table.insert(ghosts, Ghost.new(Vector.new(x, y)))
    end
end

local BOX_POWERUPS = {
    {check = box_can_heal, func = box_heal},
    {check = box_can_heal, func = box_heal},
    {check = box_can_heal, func = box_heal},
    {check = box_can_heal, func = box_heal},

    {check = box_can_triple_shot, func = box_triple_shot},
    {check = box_can_triple_shot, func = box_triple_shot},
    {check = box_can_triple_shot, func = box_triple_shot},
    {check = box_can_triple_shot, func = box_triple_shot},

    {check = box_can_exploding_shot, func = box_exploding_shot},
    {check = box_can_exploding_shot, func = box_exploding_shot},
    {check = box_can_exploding_shot, func = box_exploding_shot},
    {check = box_can_exploding_shot, func = box_exploding_shot},

    {check = box_can_invincibility, func = box_invincibility},
    {check = box_can_invincibility, func = box_invincibility},
    {check = box_can_invincibility, func = box_invincibility},

    {check = box_can_fly, func = box_fly},
    {check = box_can_fly, func = box_fly},
    {check = box_can_fly, func = box_fly},

    {check = box_can_freeze, func = box_freeze},
    {check = box_can_freeze, func = box_freeze},

    {check = box_can_smartbomb, func = box_smartbomb},
    {check = box_can_smartbomb, func = box_smartbomb},

    {check = box_can_destroy_spawn, func = box_destroy_spawn}
}

local function use_mystery_box()
    if PersistantGameData.get_num_mystery_boxes() > 0 then
        local possible_powers = {}
        for _, powerup in ipairs(BOX_POWERUPS) do
            if powerup.check() then
                table.insert(possible_powers, powerup.func)
            end
        end

        if #possible_powers > 0 then
            local p_func = Random.select_from_array(possible_powers)
            p_func()
            PersistantGameData.use_mystery_box()
        end
    end
end

----------------------------------------------------------------------

local function position_pause_menu()
    local screen_width
    local screen_height
    screen_width, screen_height = ScreenManager.dimensions()

    pause_menu.x = (screen_width - pause_menu.width) / 2
    pause_menu.y = (screen_height - pause_menu.height) / 2
end

local function mouse_data()
    local cursor = 'target'
    local x = -8
    local y = -8

    if paused then
        cursor = 'mouse'
        x = 0
        y = 0
    end

    return cursor, x, y
end
Game.mouse_data = mouse_data

local function mouse_press(x, y)
    local next_state

    if paused then
        position_pause_menu()
        if Menu.click(pause_menu, x, y) then
            menu_just_clicked = MENU_CLICKED_DELAY
        end

        if quit_game then
            PersistantGameData.end_game()
            end_game()
            next_state = Constants.STATE_TITLE
        end
    end

    return next_state
end
Game.mouse_press = mouse_press

local function mouse_release()
end
Game.mouse_release = mouse_release

local function wheel_up()
end
Game.wheel_up = wheel_up

local function wheel_down()
end
Game.wheel_down = wheel_down

local function do_pause()
    SoundManager.pause_current()

	current_ambience:pause()
	if old_ambience then
		old_ambience:pause()
	end

    Menu.play_click_sound()
    paused = true
end

local function do_keydown(key)
    if key == 'escape' then
        do_pause()
    elseif key == Settings.get('key_mystery_box') then
        use_mystery_box()
    end
end

local function keypressed(key, unicode)
    local next_state

    if key == Settings.get('key_screenshot') then
        Utility.take_screenshot()
    elseif paused then
        if key == 'escape' then
            Menu.run_option(pause_menu, #pause_menu.choices)

            if quit_game then
                PersistantGameData.end_game()
                end_game()
                next_state = Constants.STATE_TITLE
            end
        end
    elseif Actor.is_alive(player) then
        do_keydown(key)
    end

    return next_state
end
Game.keypressed = keypressed

local function keyreleased(key, unicode)
end
Game.keyreleased = keyreleased

local function set_paused(is_paused)
    if is_paused then
        do_pause()
    end
end
Game.set_paused = set_paused

----------------------------------------------------------------------

local function update_ghosts()
    local ghost_index = 1
    while ghost_index <= #ghosts do
        local g = ghosts[ghost_index]

        if g.actor_section and not spawns[g.actor_section] then
            Ghost.remove_spawn(g)
        end

        Ghost.update(g)
        if Ghost.finished(g) then
            table.remove(ghosts, ghost_index)
            if g.actor_section then
                Spawn.schedule_spawn(
                  spawns[g.actor_section], g.actor_name)
            end
        else
            ghost_index = ghost_index + 1
        end
    end
end

local function update_spawns()
    for _, s in pairs(spawns) do
        local new_monster, new_explosion = Spawn.update(s)
        if new_monster then
            table.insert(monsters, new_monster)
            table.insert(particles, new_explosion)
        end
    end
end

local function update_particles()
    local particle_index = 1
    while particle_index <= #particles do
        local p = particles[particle_index]
        Particles.update(p)

        if Particles.finished(p) then
            table.remove(particles, particle_index)
        else
            particle_index = particle_index + 1
        end
    end

    if player_flight_particles then
        Particles.update(player_flight_particles)
    end
end

local function update_bullets()
    for _, b in ipairs(bullets) do
        Bullet.update(b)
    end
end

local function update_ambient_sound()
    local required_ambience

    local is_chasm_room = player.current_room.chasm
    if is_chasm_room then
        required_ambience = chasm_ambience_sound
    else
        required_ambience = inner_ambience_sound
    end

    local master_volume = Settings.get('volume')

    if not current_ambience then
        current_ambience = required_ambience
        current_ambience:setVolume(master_volume)
        current_ambience:play()
        ambience_fade_time = 0
    elseif current_ambience ~= required_ambience then
        old_ambience = current_ambience
        current_ambience = required_ambience
        current_ambience:play()

        ambience_fade_time = AMBIENCE_FADE_TIME + 1
    end

    if ambience_fade_time > 0 then
        ambience_fade_time = ambience_fade_time - 1
        local delta = ambience_fade_time / AMBIENCE_FADE_TIME
        local inverse_delta = 1 - delta

        if ambience_fade_time == 0 then
            old_ambience:stop()
            current_ambience:setVolume(master_volume)
        else
            old_ambience:setVolume(delta * master_volume)
            current_ambience:setVolume(inverse_delta * master_volume)
        end
    end
end

local function update_current_room(act)
    local pos = Actor.pos(act)

    pos.x = math.floor(pos.x
      / (Dungeon.TOTAL_ROOM_SIZE * Constants.TILE_SIZE))
    pos.y = math.floor(pos.y
      / (Dungeon.TOTAL_ROOM_SIZE * Constants.TILE_SIZE))

    if not act.current_room or (act.current_room.x ~= pos.x)
      or (act.current_room.y ~= pos.y) then
        act.current_room = Graph.get_vertex_by_value(
          map_graph, pos, Vector.equals)
        if act == player then
            act.current_room.explored = true
        end
    end
end

local function update_player()
    local screen_width
    local screen_height
    screen_width, screen_height = ScreenManager.dimensions()

    local mouse_x
    local mouse_y
    mouse_x, mouse_y = MouseManager.get_pos()

    local player_shooting = love.mouse.isDown('l')
    if menu_just_clicked > 0 then
        player_shooting = false
        menu_just_clicked = menu_just_clicked - 1
    end

    local custom_left
    if player_shooting then
        custom_left = mouse_x < (screen_width / 2)
    end

    local move_x = 0
    local move_y = 0

    -- North
    if love.keyboard.isDown(Settings.get('key_north')) then
        move_y = move_y - 1
    end
    -- East
    if love.keyboard.isDown(Settings.get('key_east')) then
        move_x = move_x + 1
    end
    -- South
    if love.keyboard.isDown(Settings.get('key_south')) then
        move_y = move_y + 1
    end
    -- West
    if love.keyboard.isDown(Settings.get('key_west')) then
        move_x = move_x - 1
    end

    Actor.update(player)
    Actor.set_velocity(player, move_x, move_y, custom_left)

    if Actor.can_shoot(player) and player_shooting then
        local dx = mouse_x - (screen_width / 2)
        local dy = mouse_y - (screen_height / 2) + player.shoot_y

        local angle = Utility.coords_to_angle(dx, dy)

        local triple_shot = (player.triple_shots > 0)

        table.insert(bullets, Actor.shoot(
            player, current_map, angle, false, triple_shot
        ))
        if triple_shot then
            player.triple_shots = player.triple_shots - 1

            table.insert(bullets, Actor.shoot(
                player, current_map, angle + TRIPLE_SHOT_ANGLE, true
            ))
            table.insert(bullets, Actor.shoot(
                player, current_map, angle - TRIPLE_SHOT_ANGLE, true
            ))
        end
        if player.exploding_shots > 0 then
            player.exploding_shots = player.exploding_shots - 1
        end
    end

    update_current_room(player)
    update_ambient_sound()
end

local function update_monsters()
    for _, m in ipairs(monsters) do
        update_current_room(m)
        Actor.update(m)
        if (ai_delay == 0) and (m.freeze_time == 0) then
            local new_bullet = AI.run(m, map_grid, player,
              unlocked_regions, door_tiles, map_graph, current_map)
            if new_bullet then
                table.insert(bullets, new_bullet)
            end
        else
            Actor.set_velocity(m, 0, 0)
        end
    end

    AI.reset_globals()
end

local function update_physics()
    Map.update_physics(current_map)
    processed_colliders = {}
    removed_colliders = {}

    while #fixtures_to_remove > 0 do
        local f = fixtures_to_remove[1]
        table.remove(fixtures_to_remove, 1)
        f:destroy()
    end

    while #bodies_to_remove > 0 do
        local b = bodies_to_remove[1]
        table.remove(bodies_to_remove, 1)
        b:destroy()
    end
end

local function update_minimap()
    ColourFlux.update(current_room_flux)
end

local function do_update()
    local next_state

    AimGuide.update()

    if overlay_time > 0 then
        overlay_time = overlay_time - 1
        if overlay_time == 0 then
            overlay_colour = nil
        end
    end

    if ai_delay > 0 then
        ai_delay = ai_delay - 1
    end

    update_ghosts()
    update_spawns()
    update_particles()
    update_bullets()

    update_monsters()

    if not Actor.is_alive(player) then
        game_over_timer = game_over_timer - 1
        if game_over_timer == 0 then
            PersistantGameData.end_game()
            end_game()
            next_state = Constants.STATE_TITLE
        end
    else
        update_player()
    end

    if not next_state then
        update_physics()

        -- Bullet explosions
        while #bullet_explosions > 0 do
            local e = bullet_explosions[1]
            table.remove(bullet_explosions, 1)

            local new_bullets = Bullet.create_bullet_explosion(
                e.pos, e.actor, current_map, e.actors_to_ignore
            )
            for _, b in ipairs(new_bullets) do
                table.insert(bullets, b)
            end
        end

        if exit_found then
            PersistantGameData.set_player_data(
                player.health,
                player.triple_shots,
                player.exploding_shots,
                player.invincibility_time
            )
            end_game()
            next_state = Constants.STATE_LEVEL_WARP
        elseif Actor.is_alive(player) then
            local player_pos_i = player_pos
            player_pos = Actor.pos(player)

            local delta = Vector.subtract(player_pos, player_pos_i)
            if (delta.x ~= 0) or (delta.y ~= 0)
              then
                Parallax.move(abyss, delta.x, delta.y)
            end

            SoundManager.set_listener(player_pos.x, player_pos.y)
            update_minimap()
        end
    end

    return next_state
end

local function update()
    local next_state

    if paused then
        Menu.update(pause_menu)
    else
        next_state = do_update()
    end

    return next_state
end
Game.update = update

----------------------------------------------------------------------

local function position_camera(screen_width, screen_height)
    local half_map_width
      = math.floor(screen_width / (Constants.TILE_SIZE * 2)) + 1
    local half_map_height
      = math.floor(screen_height / (Constants.TILE_SIZE * 2)) + 1

    local tile_x = math.floor(player_pos.x / Constants.TILE_SIZE)
    local tile_y = math.floor(player_pos.y / Constants.TILE_SIZE)

    love.graphics.translate(
      (screen_width / 2) - player_pos.x,
      (screen_height / 2) - player_pos.y
    )

    return tile_x, tile_y, half_map_width, half_map_height
end

local function draw_world(screen_width, screen_height)
    love.graphics.push()

    local tile_x
    local tile_y
    local half_map_width
    local half_map_height

    tile_x, tile_y, half_map_width, half_map_height
      = position_camera(screen_width, screen_height)

    Map.draw_bottom(
      current_map,
      tile_x - half_map_width, tile_y - half_map_height,
      tile_x + half_map_width, tile_y + half_map_height
    )

    local drawables = {}

    local entrance_drawables = Entrance.entrance_draw_data(entrance)
    for _, d in ipairs(entrance_drawables) do
        table.insert(drawables, d)
    end

    local exit_drawables = Entrance.exit_draw_data(exit)
    for _, d in ipairs(exit_drawables) do
        table.insert(drawables, d)
    end

    for _, d in ipairs(doors) do
        table.insert(drawables, Door.draw_data(d))
    end
    for _, k in pairs(keys) do
        table.insert(drawables, Key.draw_data(k))
    end
    for _, b in ipairs(mystery_boxes) do
        table.insert(drawables, MysteryBox.draw_data(b))
    end

    for _, s in pairs(spawns) do
        local spawn_drawables = Spawn.draw_data(s)
        for _, d in ipairs(spawn_drawables) do
            table.insert(drawables, d)
        end
    end

    if Actor.is_alive(player) then
        local plr_draw_data = Actor.draw_data(player)
        for _, plrdata in ipairs(plr_draw_data) do
            table.insert(drawables, plrdata)
        end
    end

    for _, m in ipairs(monsters) do
        local m_draw_data = Actor.draw_data(m)
        for _, mdraw in ipairs(m_draw_data) do
            table.insert(drawables, mdraw)
        end
    end

    for _, b in ipairs(bullets) do
        table.insert(drawables, Bullet.draw_data(b))
    end

    table.sort(drawables, function(a, b)
        return (a.z < b.z) or ((a.z == b.z) and (a.real_y < b.real_y))
    end)

    for _, d in ipairs(drawables) do
        if d.colour then
            love.graphics.setColor(d.colour)
        end

        if d.is_rectangle then
            love.graphics.rectangle('fill', d.x, d.y,
              d.width, d.height)
        else
            love.graphics.drawq(d.img, d.quad, d.x, d.y)
        end

        love.graphics.setColor(255, 255, 255, 255)
    end

    Map.draw_top(
      current_map,
      tile_x - half_map_width, tile_y - half_map_height,
      tile_x + half_map_width, tile_y + half_map_height
    )

    for _, g in ipairs(ghosts) do
        Ghost.draw(g)
    end

    for _, p in ipairs(particles) do
        Particles.draw(p)
    end

    if player_flight_particles then
        love.graphics.draw(player_flight_particles,
          player_pos.x, player_pos.y)
    end

    for _, s in pairs(spawns) do
        Spawn.draw_particles(s)
    end

    -- Reset camera
    love.graphics.pop()
end

local function draw_gui(screen_width, screen_height)
    if Actor.is_alive(player) and not paused then
        AimGuide.draw()
    end

    love.graphics.setFont(gui_font)

    -- Keys

    love.graphics.setColor(255, 255, 255, GUI_ALPHA)

    local key_y = screen_height - Key.FULL_HEIGHT - GUI_MARGIN
    for i, key_index in ipairs(collected_keys) do
        x = screen_width - Key.FULL_WIDTH - GUI_MARGIN
          - ((i - 1) * (Key.FULL_WIDTH + GUI_INNER_MARGIN))
        Key.draw_sprite(key_index, x, key_y)
    end

    -- Healthbar

    local percentage = math.max(0, Actor.health_percentage(player))

    local bar_x = GUI_MARGIN
    local bar_y = screen_height - HEALTHBAR_HEIGHT - GUI_MARGIN
      - HEALTHBAR_Y_ADD

    local bar_width = screen_width - GUI_MARGIN - KEY_AREA_WIDTH
    local invincible_bar_width
      = (player.invincibility_time / INVINCIBILITY_TIME) * bar_width
    local filled_bar_width = bar_width * percentage
    local empty_bar_width

    if invincible_bar_width > filled_bar_width then
        filled_bar_width = 0
        empty_bar_width = bar_width - invincible_bar_width
    else
        empty_bar_width = bar_width - filled_bar_width
        filled_bar_width = filled_bar_width - invincible_bar_width
    end

    love.graphics.setColor(HEALTHBAR_FILL_INVINCIBLE_COLOUR)
    love.graphics.rectangle(
        'fill', bar_x, bar_y, invincible_bar_width, HEALTHBAR_HEIGHT
    )

    love.graphics.setColor(HEALTHBAR_FILL_COLOUR)
    love.graphics.rectangle(
        'fill', bar_x + invincible_bar_width, bar_y,
        filled_bar_width, HEALTHBAR_HEIGHT
    )

    love.graphics.setColor(HEALTHBAR_BACK_COLOUR)
    love.graphics.rectangle(
        'fill',
        bar_x + invincible_bar_width + filled_bar_width, bar_y,
        empty_bar_width, HEALTHBAR_HEIGHT
    )

    love.graphics.setColor(255, 255, 255, GUI_ALPHA)

    local icon_y = bar_y

    -- Mystery boxes

    if PersistantGameData.get_num_mystery_boxes() > 0 then
        icon_y = bar_y - MysteryBox.HEIGHT - GUI_INNER_MARGIN

        MysteryBox.draw_sprite(GUI_MARGIN, icon_y)

        local box_text_x = GUI_MARGIN + MysteryBox.WIDTH
          + GUI_INNER_MARGIN
        assert(MysteryBox.HEIGHT >= gui_font_height)
        local box_text_y = icon_y
          + (MysteryBox.HEIGHT - gui_font_height) / 2
        love.graphics.print(
          PersistantGameData.get_num_mystery_boxes(),
          box_text_x, box_text_y
        )
    end

    -- Triple shots

    if player.triple_shots > 0 then
        icon_y = icon_y - icon_triple_shot:getHeight()
          - GUI_INNER_MARGIN
        love.graphics.draw(icon_triple_shot, GUI_MARGIN, icon_y)

        local text_x = GUI_MARGIN + icon_triple_shot:getWidth()
          + GUI_INNER_MARGIN
        love.graphics.print(player.triple_shots, text_x, icon_y)
    end

    -- Exploding shots

    if player.exploding_shots > 0 then
        icon_y = icon_y - icon_exploding_shot:getHeight()
          - GUI_INNER_MARGIN
        love.graphics.draw(icon_exploding_shot, GUI_MARGIN, icon_y)

        local text_x = GUI_MARGIN + icon_exploding_shot:getWidth()
          + GUI_INNER_MARGIN
        love.graphics.print(player.exploding_shots, text_x, icon_y)
    end

    -- Minimap

    for room, _ in pairs(map_graph.vertices) do
        if room.explored then
            local x = (room.x * MINIMAP_ROOM_SPACE)
              - MINIMAP_ROOM_MARGIN + GUI_MARGIN
            local y = (room.y * MINIMAP_ROOM_SPACE)
              - MINIMAP_ROOM_MARGIN + GUI_MARGIN

            if room == player.current_room then
                room_colour = {
                    ColourFlux.get_colour(current_room_flux)
                }
            else
                room_colour = MINIMAP_ROOM_COLOUR
            end

            love.graphics.setColor(room_colour)

            local room_margin = MINIMAP_ROOM_MARGIN
            local room_size = MINIMAP_ROOM_SIZE

            if room.corridor then
                room_size = MINIMAP_DOOR_SIZE
                room_margin = MINIMAP_DOOR_BORDER
            end

            love.graphics.rectangle('fill',
              x + room_margin, y + room_margin, room_size, room_size)

            if room.spawn and spawns[room.section] then
                love.graphics.setColor(128, 255, 255, GUI_ALPHA)

                love.graphics.circle('fill',
                  x + MINIMAP_ROOM_MIDDLE,
                  y + MINIMAP_ROOM_MIDDLE,
                  MINIMAP_SPAWN_RADIUS, 4)
            end

            if room.exit then
                love.graphics.setColor(0, 81, 211, GUI_ALPHA)

                love.graphics.circle('fill',
                  x + MINIMAP_ROOM_MIDDLE,
                  y + MINIMAP_ROOM_MIDDLE,
                  MINIMAP_EXIT_RADIUS)

                love.graphics.setColor(0, 54, 118, GUI_ALPHA)

                love.graphics.circle('fill',
                  x + MINIMAP_ROOM_MIDDLE,
                  y + MINIMAP_ROOM_MIDDLE,
                  MINIMAP_EXIT_RADIUS / 2)
            end

            local neighbors = Graph.get_adjacent(map_graph, room)
            for _, adj in ipairs(neighbors) do
                local dir, delta = Vector.get_dir(room, adj)

                local door_x
                local door_y
                local door_width
                local door_height

                local door_border = MINIMAP_DOOR_BORDER
                local door_size = MINIMAP_DOOR_SIZE

                local edge = Graph.get_edge(map_graph, room, adj)

                if edge.join then
                    door_border
                      = (MINIMAP_ROOM_SPACE - MINIMAP_ROOM_SIZE) / 2
                    door_size = MINIMAP_ROOM_SIZE
                end

                if dir == 'north' then
                    door_x = x + door_border
                    door_y = y

                    door_width = door_size
                    door_height = MINIMAP_ROOM_MARGIN

                    if room.corridor then
                        love.graphics.setColor(room_colour)
                        love.graphics.rectangle('fill',
                          door_x, door_y + door_height,
                          door_width, room_margin - door_height)
                    end
                elseif dir == 'east' then
                    door_x = x + MINIMAP_ROOM_SIZE
                      + MINIMAP_ROOM_MARGIN
                    door_y = y + door_border

                    door_width = MINIMAP_ROOM_MARGIN
                    door_height = door_size

                    if room.corridor then
                        love.graphics.setColor(room_colour)
                        love.graphics.rectangle('fill',
                          door_x - door_width - 1, door_y,
                          room_margin - door_width, door_height)
                    end
                elseif dir == 'south' then
                    door_x = x + door_border
                    door_y = y + MINIMAP_ROOM_SIZE
                      + MINIMAP_ROOM_MARGIN

                    door_width = door_size
                    door_height = MINIMAP_ROOM_MARGIN

                    if room.corridor then
                        love.graphics.setColor(room_colour)
                        love.graphics.rectangle('fill',
                          door_x, door_y - door_height - 1,
                          door_width, room_margin - door_height)
                    end
                else
                    assert(dir == 'west')
                    door_x = x
                    door_y = y + door_border

                    door_width = MINIMAP_ROOM_MARGIN
                    door_height = door_size

                    if room.corridor then
                        love.graphics.setColor(room_colour)
                        love.graphics.rectangle('fill',
                          door_x + door_width, door_y,
                          room_margin - door_width, door_height)
                    end
                end

                if edge.door then
                    local c = Constants.DOOR_COLOURS[edge.door]
                    love.graphics.setColor(
                      c[1], c[2], c[3], GUI_ALPHA
                    )
                else
                    love.graphics.setColor(room_colour)
                end

                love.graphics.rectangle('fill', door_x, door_y,
                  door_width, door_height)
            end
        end
    end

    -- Current level

    local text = 'Structure ' .. PersistantGameData.current_level()
    local text_width = gui_font:getWidth(text)
    local text_x = screen_width - text_width - GUI_MARGIN

    love.graphics.setColor(255, 255, 255, GUI_ALPHA)
    love.graphics.print(text, text_x, GUI_MARGIN)

    love.graphics.setColor(255, 255, 255, 255)
end

local function draw()
    local screen_width
    local screen_height
    screen_width, screen_height = ScreenManager.dimensions()

    Parallax.draw(abyss)
    draw_world(screen_width, screen_height)
    draw_gui(screen_width, screen_height)

    if overlay_time > 0 then
        assert(overlay_colour)
        alpha = 128 * (overlay_time / OVERLAY_TIME)

        love.graphics.setColor(
          overlay_colour[1], overlay_colour[2], overlay_colour[3],
          alpha
        )
        love.graphics.rectangle(
            'fill', 0, 0, screen_width, screen_height
        )

        love.graphics.setColor(255, 255, 255, 255)
    end

    if paused then
        love.graphics.setColor(0, 0, 0, 64)
        love.graphics.rectangle(
            'fill', 0, 0, screen_width, screen_height
        )
        love.graphics.setColor(255, 255, 255, 255)

        position_pause_menu()
        Menu.draw(pause_menu)
    end
end
Game.draw = draw

----------------------------------------------------------------------

return Game