--
-- Copyright 2012 Aaron MacDonald
--
-- Licensed under the Apache License, Version 2.0 (the "License");
-- you may not use this file except in compliance with the License.
-- You may obtain a copy of the License at
--
--     http://www.apache.org/licenses/LICENSE-2.0
--
-- Unless required by applicable law or agreed to in writing, software
-- distributed under the License is distributed on an "AS IS" BASIS,
-- WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
-- implied.
-- See the License for the specific language governing permissions and
-- limitations under the License.
--

local Ghost = {}

local Constants = require 'constants'
local Utility = require 'utility'
local Vector = require 'vector'

local Actor = require 'actor'

----------------------------------------------------------------------

local SPEED = 500
local DIST_PER_FRAME = SPEED * Constants.SECONDS_PER_FRAME

local EYES_IMG = 'eyes'

local EYES_WIDTH = 24
local EYES_HEIGHT = 20

local TRANS_X = EYES_WIDTH / 2
local TRANS_Y = EYES_HEIGHT

local MAP_PADDING = 128

local eyes_image
local eyes_quad

----------------------------------------------------------------------

local function velocity_and_frames(pos, dest_pos, no_spawn)
    local delta = Vector.subtract(dest_pos, pos)
    local dist = Vector.length(delta)

    local velocity = Vector.normal(delta)

    local dist_per_frame = DIST_PER_FRAME
    if no_spawn then
        dist_per_frame = dist_per_frame / 2
    end

    Vector.scale_inline(velocity, dist_per_frame)

    local frames = math.ceil(dist / dist_per_frame)

    return velocity, frames
end

local function new(pos, actor_name, spawn)
    if not eyes_image then
        eyes_image = love.graphics.newImage(
          Utility.img_path(EYES_IMG)
        )
        eyes_quad = love.graphics.newQuad(
          0, 0, EYES_WIDTH, EYES_HEIGHT,
          eyes_image:getWidth(), eyes_image:getHeight()
        )
    end

    local dest_pos
    local section
    if spawn then
        dest_pos = spawn.pos
        section = spawn.section
    else
        dest_pos = Vector.new(pos.x, -MAP_PADDING)
    end

    local velocity
    local frames
    velocity, frames = velocity_and_frames(pos, dest_pos, not spawn)

    return {
        actor_name = actor_name,
        actor_section = section,
        pos = pos,
        velocity = velocity,
        frames_left = frames
    }
end
Ghost.new = new

local function remove_spawn(ghost)
    ghost.actor_section = nil

    local velocity
    local frames
    velocity, frames = velocity_and_frames(
        ghost.pos, Vector.new(ghost.pos.x, -MAP_PADDING), true
    )

    ghost.velocity = velocity
    ghost.frames = frames
    ghost.frames_left = frames
end
Ghost.remove_spawn = remove_spawn

local function finished(ghost)
    return ghost.frames_left == 0
end
Ghost.finished = finished

local function update(ghost)
    ghost.pos = Vector.add(ghost.pos, ghost.velocity)
    ghost.frames_left = ghost.frames_left - 1
end
Ghost.update = update

local function draw(ghost)
    local x = ghost.pos.x - TRANS_X
    local y = ghost.pos.y - TRANS_Y
    love.graphics.drawq(eyes_image, eyes_quad, x, y)
end
Ghost.draw = draw

----------------------------------------------------------------------

return Ghost