--
-- Copyright 2012 Aaron MacDonald
--
-- Licensed under the Apache License, Version 2.0 (the "License");
-- you may not use this file except in compliance with the License.
-- You may obtain a copy of the License at
--
--     http://www.apache.org/licenses/LICENSE-2.0
--
-- Unless required by applicable law or agreed to in writing, software
-- distributed under the License is distributed on an "AS IS" BASIS,
-- WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
-- implied.
-- See the License for the specific language governing permissions and
-- limitations under the License.
--

local Graph = {}

local Utility = require 'utility'

----------------------------------------------------------------------

local function new()
    return {
        vertices = {},
        adjacencies = {}
    }
end
Graph.new = new

local function add_vertex(gr, vert)
    gr.vertices[vert] = true
    gr.adjacencies[vert] = {}
end
Graph.add_vertex = add_vertex

local function get_vertex_by_value(gr, val, equ)
    return Utility.find_key(gr.vertices, val, equ)
end
Graph.get_vertex_by_value = get_vertex_by_value

local function set_edge(gr, vert_a, vert_b, edge)
    if not edge then
        edge = true
    end

    gr.adjacencies[vert_a][vert_b] = edge
    gr.adjacencies[vert_b][vert_a] = edge
end
Graph.set_edge = set_edge

local function get_edge(gr, vert_a, vert_b)
    local result = nil

    if gr.adjacencies[vert_a] then
        result = gr.adjacencies[vert_a][vert_b]
    end

    return result
end
Graph.get_edge = get_edge

local function are_adjacent(gr, vert_a, vert_b)
    return gr.adjacencies[vert_a][vert_b] ~= nil
end
Graph.are_adjacent = are_adjacent

local function get_adjacent(gr, vert)
    return Utility.keys_to_array(gr.adjacencies[vert])
end
Graph.get_adjacent = get_adjacent

local function get_all_edges(gr)
    local result = {}
    local half_adj = {}

    for ra, adj in pairs(gr.adjacencies) do
        for rb, e in pairs(adj) do
            if not (half_adj[rb] and half_adj[rb][ra]) then
                Utility.set_double_key(half_adj, ra, rb, e)
            end
        end
    end

    for ra, adj in pairs(half_adj) do
        for rb, e in pairs(adj) do
            table.insert(result, {ra, rb, e})
        end
    end

    return result
end
Graph.get_all_edges = get_all_edges

local function remove_edge(gr, vert_a, vert_b)
    assert(gr.vertices[vert_a])
    assert(gr.vertices[vert_b])
    assert(gr.adjacencies[vert_a] == vert_b)

    gr.adjacencies[vert_a][vert_b] = nil
    gr.adjacencies[vert_b][vert_a] = nil
end
Graph.remove_edge = remove_edge

local function remove_vertex(gr, vert)
    local adj = get_adjacent(gr, vert)
    for _, a in ipairs(adj) do
        remove_edge(vert, a)
    end
    gr.vertices[vert] = nil
end
Graph.remove_vertex = remove_vertex

----------------------------------------------------------------------

local function dfs(gr, vert, explored, parent, visit_func)
    visit_func(vert, parent)
    explored[vert] = true

    local adj = get_adjacent(gr, vert)
    for _, a in ipairs(adj) do
        if not explored[a] then
            dfs(gr, a, explored, vert, visit_func)
        end
    end
end

local function depth_first_search(gr, start_vert, visit_func)
    local explored = {}
    dfs(gr, start_vert, explored, nil, visit_func)
end
Graph.depth_first_search = depth_first_search

----------------------------------------------------------------------

return Graph