--
-- Copyright 2012 Aaron MacDonald
--
-- Licensed under the Apache License, Version 2.0 (the "License");
-- you may not use this file except in compliance with the License.
-- You may obtain a copy of the License at
--
--     http://www.apache.org/licenses/LICENSE-2.0
--
-- Unless required by applicable law or agreed to in writing, software
-- distributed under the License is distributed on an "AS IS" BASIS,
-- WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
-- implied.
-- See the License for the specific language governing permissions and
-- limitations under the License.
--

local Grid = {}

local Vector = require 'vector'
local Utility = require 'utility'

----------------------------------------------------------------------

local function new(width, height)
    return {
        cells = {},
        width = width,
        height = height
    }
end
Grid.new = new

local function in_bounds(gr, x, y)
    return not gr.width or ((x >= 0) and (y >= 0)
      and (x < gr.width) and (y < gr.height))
end
Grid.in_bounds = in_bounds

local function set(gr, x, y, data)
    Utility.set_double_key(gr.cells, x, y, data)
end
Grid.set = set

local function get(gr, x, y)
    local result = nil
    if gr.cells[x] then
        result = gr.cells[x][y]
    end
    return result
end
Grid.get = get

----------------------------------------------------------------------

local function flood_fill(gr, start_x, start_y,
  min_x, min_y, max_x, max_y, visit_f, is_passable_f)
    if not min_x then
        min_x = 0
    end
    if not min_y then
        min_y = 0
    end
    if not max_x then
        max_x = gr.width
    end
    if not max_y then
        max_y = gr.height
    end

    assert(max_x, max_x)
    assert(max_y, max_y)
    assert(start_x >= min_x, start_x .. ', ' .. min_x)
    assert(start_y >= min_y, start_y .. ', ' .. min_y)
    assert(start_x <= max_x, start_x .. ', ' .. max_x)
    assert(start_y <= max_y, start_y .. ', ' .. max_y)
    assert(get(gr, start_x, start_y))

    local visited = {}
    local frontier = { Vector.new(start_x, start_y) }

    while #frontier > 0 do
        local cell = frontier[#frontier]
        frontier[#frontier] = nil

        if not visited[cell.x] or not visited[cell.x][cell.y] then
            visit_f(cell.x, cell.y)

            Utility.set_double_key(visited, cell.x, cell.y, true)

            for _, a in ipairs(Vector.cardinal_adjacent(cell)) do
                local passable = not is_passable_f
                  or is_passable_f(a.x, a.y)
                local in_bounds = (a.x >= min_x) and (a.y >= min_y)
                  and (a.x <= max_x) and (a.y <= max_y)

                if passable and in_bounds then
                    table.insert(frontier, a)
                end
            end
        end
    end
end
Grid.flood_fill = flood_fill

----------------------------------------------------------------------

return Grid