--
-- Copyright 2012 Aaron MacDonald
--
-- Licensed under the Apache License, Version 2.0 (the "License");
-- you may not use this file except in compliance with the License.
-- You may obtain a copy of the License at
--
--     http://www.apache.org/licenses/LICENSE-2.0
--
-- Unless required by applicable law or agreed to in writing, software
-- distributed under the License is distributed on an "AS IS" BASIS,
-- WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
-- implied.
-- See the License for the specific language governing permissions and
-- limitations under the License.
--

local LevelWarp = {}

local Settings = require 'settings'
local Constants = require 'constants'
local Utility = require 'utility'

local ScreenManager = require 'screen_manager'
local SoundManager = require 'sound_manager'

----------------------------------------------------------------------

local TIME_LENGTH = 45
local current_time

local NUM_CIRCLES = 15

local CIRCLE_ORIGIN_Y = 0.2 -- Percentage
local CIRCLE_DY = 2

local CIRCLE_ORIGIN_RADIUS = 0

local CIRCLE_ORIGIN_LINE_WIDTH = 2
local CIRCLE_DELTA_LINE_WIDTH = 0.05

local WARP_SOUND = 'warp'

local PLAYER_SPRITE_NAME = 'player_running'

local player_img
local player_half_width
local player_half_height
local player_quarter_height

----------------------------------------------------------------------

local function load()
    player_img = love.graphics.newImage(
      Utility.img_path(PLAYER_SPRITE_NAME))
    player_half_width = player_img:getWidth() / 2
    player_half_height = player_img:getHeight() / 2
    player_quarter_height = player_img:getHeight() / 4
end
LevelWarp.load = load

local function init()
    current_time = 0
    SoundManager.new_sound(WARP_SOUND)
end
LevelWarp.init = init

local function end_warp()
    current_time = 0
end

----------------------------------------------------------------------

local function mouse_data()
    return nil
end
LevelWarp.mouse_data = mouse_data

local function mouse_press()
end
LevelWarp.mouse_press = mouse_press

local function mouse_release()
end
LevelWarp.mouse_release = mouse_release

local function wheel_up()
end
LevelWarp.wheel_up = wheel_up

local function wheel_down()
end
LevelWarp.wheel_down = wheel_down

local function keypressed(key, unicode)
    if key == Settings.get('key_screenshot') then
        Utility.take_screenshot()
    end
end
LevelWarp.keypressed = keypressed

local function keyreleased(key, unicode)
end
LevelWarp.keyreleased = keyreleased

local function set_paused(is_paused)
end
LevelWarp.set_paused = set_paused

----------------------------------------------------------------------

local function update(key, unicode)
    local next_state

    current_time = current_time + 1
    if current_time >= TIME_LENGTH then
        end_warp()
        next_state = Constants.STATE_GAME
    end

    return next_state
end
LevelWarp.update = update

----------------------------------------------------------------------

local function draw_tunnel(screen_width, screen_height)
    local origin_x = screen_width / 2
    local origin_y = screen_height * CIRCLE_ORIGIN_Y

    local end_radius = screen_width / 2
    if screen_height > screen_width then
        end_radius = screen_height / 2
    end

    local circle_drop_time = (screen_height - origin_y) / CIRCLE_DY

    local circle_delta_radius
      = (end_radius - CIRCLE_ORIGIN_RADIUS) / circle_drop_time

    for i = 0, NUM_CIRCLES do
        local y_add = (screen_height - origin_y) / NUM_CIRCLES

        local real_y = (((CIRCLE_DY * (current_time)) + (y_add * i))
          % (screen_height - origin_y)) + origin_y

        local radius = CIRCLE_ORIGIN_RADIUS
          + (circle_delta_radius * (real_y - origin_y))

        local line_width = CIRCLE_ORIGIN_LINE_WIDTH
          + (CIRCLE_DELTA_LINE_WIDTH * (real_y - origin_y))

        local cr, cg, cb = Utility.interpolate_colour(
          0, 81, 211, 255,
          255, 255, 255, 255,
          (real_y - origin_y) / (screen_height - origin_y)
        )

        love.graphics.setColor(cr, cg, cb)
        love.graphics.setLineWidth(line_width)

        love.graphics.circle('line', origin_x, real_y,
          radius, 64)

        love.graphics.setLineWidth(1)
        love.graphics.setColor(255, 255, 255)
    end

    return origin_x, origin_y
end

local function draw_player(screen_width, screen_height,
  origin_x, origin_y)
    local player_scale = Utility.interpolate(
        screen_width / player_img:getHeight(),
        0,
        current_time / TIME_LENGTH
    )
    local player_x = origin_x - (player_half_width * player_scale)
    local player_y = Utility.interpolate(
        screen_height - (player_quarter_height * player_scale),
        origin_y - (player_half_height * player_scale),
        current_time / TIME_LENGTH
    )

    love.graphics.draw(
        player_img,
        player_x, player_y,
        0,
        player_scale, player_scale
    )
end

local function draw()
    local screen_width
    local screen_height
    screen_width, screen_height = ScreenManager.dimensions()

    local origin_x
    local origin_y
    origin_x, origin_y = draw_tunnel(screen_width, screen_height)
    draw_player(screen_width, screen_height, origin_x, origin_y)
end
LevelWarp.draw = draw

----------------------------------------------------------------------

return LevelWarp