--
-- Copyright 2012 Aaron MacDonald
--
-- Licensed under the Apache License, Version 2.0 (the "License");
-- you may not use this file except in compliance with the License.
-- You may obtain a copy of the License at
--
--     http://www.apache.org/licenses/LICENSE-2.0
--
-- Unless required by applicable law or agreed to in writing, software
-- distributed under the License is distributed on an "AS IS" BASIS,
-- WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
-- implied.
-- See the License for the specific language governing permissions and
-- limitations under the License.
--

local Random = require 'random'
local Constants = require 'constants'
local Utility = require 'utility'

local Settings = require 'settings'

local MouseManager = require 'mouse_manager'
local ScreenManager = require 'screen_manager'
local SoundManager = require 'sound_manager'

local accumulated_time
local is_paused

local ICON_FILE = 'icon'
local icon

----------------------------------------------------------------------

local states = {
    [Constants.STATE_TITLE] = require 'title',
    [Constants.STATE_LEVEL_WARP] = require 'level_warp',
    [Constants.STATE_GAME] = require 'game'
}
local current_state

local function state_transition(new_state)
    if new_state and new_state ~= current_state then
        current_state = new_state
        states[current_state].init()
    end
end

----------------------------------------------------------------------

function love.load()
    love.filesystem.setIdentity('Warp Run')

    SoundManager.load()

    Settings.load()
    love.graphics.setMode(
        Settings.get('screen_width'),
        Settings.get('screen_height'),
        Settings.get('fullscreen')
    )

    love.graphics.setDefaultImageFilter('nearest', 'nearest')
    love.mouse.setVisible(false)

    icon = love.graphics.newImage(Utility.img_path(ICON_FILE))
    love.graphics.setIcon(icon)

    accumulated_time = 0
    is_paused = false

    love.physics.setMeter(Constants.TILE_SIZE)

    Random.init()
    MouseManager.load()

    for _, state in pairs(states) do
        state.load()
    end

    state_transition(Constants.STATE_TITLE)
end

function love.focus(has_focus)
    is_paused = not has_focus
    states[current_state].set_paused(is_paused)
end

----------------------------------------------------------------------

function love.mousepressed(x, y, button)
    local scale = ScreenManager.get_scale()
    local mx = x * scale
    local my = y * scale

    if button == 'l' then
        next_state = states[current_state].mouse_press(mx, my)
    elseif button == 'wu' then
        next_state = states[current_state].wheel_up(mx, my)
    elseif button == 'wd' then
        next_state = states[current_state].wheel_down(mx, my)
    end

    state_transition(next_state)
end

function love.mousereleased(x, y, button)
    if button == 'l' then
        local scale = ScreenManager.get_scale()
        local next_state = states[current_state].mouse_release(
          x * scale, y * scale)
        state_transition(next_state)
    end
end

function love.keypressed(key, unicode)
    local next_state = states[current_state].keypressed(
      key, unicode)
    state_transition(next_state)
end

function love.keyreleased(key, unicode)
    local next_state = states[current_state].keyreleased(
      key, unicode)
    state_transition(next_state)
end

----------------------------------------------------------------------

function love.update(dt)
    accumulated_time = accumulated_time + dt
    while accumulated_time >= Constants.SECONDS_PER_FRAME do
        MouseManager.update()
        if not is_paused then
            local next_state = states[current_state].update()
            SoundManager.update()
            state_transition(next_state)
        end
        accumulated_time
          = accumulated_time - Constants.SECONDS_PER_FRAME
    end

    local caption = 'Warp Run - ' .. love.timer.getFPS() .. ' FPS'
    love.graphics.setCaption(caption)
end

----------------------------------------------------------------------

function love.draw()
    ScreenManager.scale_screen()

    states[current_state].draw()
    MouseManager.draw(states[current_state].mouse_data())
end