--
-- Copyright 2012 Aaron MacDonald
--
-- Licensed under the Apache License, Version 2.0 (the "License");
-- you may not use this file except in compliance with the License.
-- You may obtain a copy of the License at
--
--     http://www.apache.org/licenses/LICENSE-2.0
--
-- Unless required by applicable law or agreed to in writing, software
-- distributed under the License is distributed on an "AS IS" BASIS,
-- WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
-- implied.
-- See the License for the specific language governing permissions and
-- limitations under the License.
--

local Map = {}

local Grid = require 'grid'
local TileSet = require 'tileset'
local Constants = require 'constants'

----------------------------------------------------------------------

local function new(width, height)
    local phys_world = love.physics.newWorld(0, 0)

    local body = love.physics.newBody(phys_world, 0, 0)

    return {
        width = width,
        height = height,

        tilesets = {},

        layers = {},
        layer_offsets = {},
        object_index = 0,
        max_layer = 0,

        phys_world = phys_world,
        phys_body = body,
        wall_shapes = {},
        wall_fixtures = {}
    }
end
Map.new = new

local function set_collision_callbacks(map, beginContact)
    map.phys_world:setCallbacks(beginContact)
end
Map.set_collision_callbacks = set_collision_callbacks

local function add_tileset(map, tilesheet_name)
    if not map.default_tileset then
        map.default_tileset = tilesheet_name
    end

    if not map.tilesets[tilesheet_name] then
        map.tilesets[tilesheet_name] = TileSet.new(tilesheet_name)
    end
end
Map.add_tileset = add_tileset

-- Layer that all objects and actors are drawn on top of
local function set_object_layer(map, index)
    map.object_index = index
end
Map.set_object_layer = set_object_layer

local function set_layer_offset(map, index, offset)
    map.layer_offsets[index] = offset
end
Map.set_layer_offset = set_layer_offset

local function set_tile(map, x, y, tile_name, layer, tileset_name)
    if not tileset_name then
        tileset_name = map.default_tileset
    end
    if not layer then
        layer = 0
    end
    if layer > map.max_layer then
        map.max_layer = layer
    end

    if not map.layers[layer] then
        map.layers[layer] = Grid.new(map.width, map.height)
    end

    Grid.set(map.layers[layer], x, y,
      {tileset = tileset_name, tile_name = tile_name})
end
Map.set_tile = set_tile

local function generate_walls(map, is_wall_f, category)
    local rectangles = {}

    for x = 0, map.width - 1 do
        local start_y
        local end_y

        for y = 0, map.height - 1 do
            if is_wall_f(x, y) then
                if not start_y then
                    start_y = y
                end
                end_y = y
            elseif start_y then
                local overlaps = {}
                for _, r in ipairs(rectangles) do
                    if (r.end_x == x - 1)
                      and (start_y <= r.start_y)
                      and (end_y >= r.end_y) then
                        table.insert(overlaps, r)
                    end
                end
                table.sort(
                    overlaps,
                    function (a, b)
                        return a.start_y < b.start_y
                    end
                )

                for _, r in ipairs(overlaps) do
                    if start_y < r.start_y then
                        local new_rect = {
                            start_x = x,
                            start_y = start_y,
                            end_x = x,
                            end_y = r.start_y - 1
                        }
                        table.insert(rectangles, new_rect)
                        start_y = r.start_y
                    end

                    if start_y == r.start_y then
                        r.end_x = r.end_x + 1

                        if end_y == r.end_y then
                            start_y = nil
                            end_y = nil
                        elseif end_y > r.end_y then
                            start_y = r.end_y + 1
                        end
                    end
                end

                if start_y then
                    local new_rect = {
                        start_x = x,
                        start_y = start_y,
                        end_x = x,
                        end_y = end_y
                    }
                    table.insert(rectangles, new_rect)

                    start_y = nil
                    end_y = nil
                end
            end
        end

        if start_y then
            local new_rect = {
                start_x = x,
                start_y = start_y,
                end_x = x,
                end_y = end_y
            }
            table.insert(rectangles, new_rect)

            start_y = nil
            end_y = nil
        end
    end

    for _, r in ipairs(rectangles) do
        local start_x = r.start_x * Constants.TILE_SIZE
        local start_y = r.start_y * Constants.TILE_SIZE
        local width = (r.end_x - r.start_x + 1) * Constants.TILE_SIZE
        local height = (r.end_y - r.start_y + 1) * Constants.TILE_SIZE

        local x = start_x + (width / 2)
        local y = start_y + (height / 2)

        local shape = love.physics.newRectangleShape(
          x, y, width, height)

        local fixture = love.physics.newFixture(map.phys_body, shape)
        fixture:setUserData({
            type = 'wall',
            data = {}
        })
        fixture:setFriction(0)

        if category then
            fixture:setCategory(category)
        end

        table.insert(map.wall_shapes, shape)
        table.insert(map.wall_fixtures, fixture)
    end
end
Map.generate_walls = generate_walls

----------------------------------------------------------------------

local function update_physics(map)
    map.phys_world:update(Constants.SECONDS_PER_FRAME)
end
Map.update_physics = update_physics

----------------------------------------------------------------------

local function draw_tile(map, layer, x, y)
    local tile = Grid.get(map.layers[layer], x, y)
    if tile then
        local tileset = map.tilesets[tile.tileset]

        local off_x = 0
        local off_y = 0
        if map.layer_offsets[layer] then
            off_x = map.layer_offsets[layer].x
            off_y = map.layer_offsets[layer].y
        end

        local dx = (x * Constants.TILE_SIZE) + off_x
        local dy = (y * Constants.TILE_SIZE) + off_y

        TileSet.draw(tileset, tile.tile_name, dx, dy)
    end
end

local function draw_layer(map, index, min_x, min_y, max_x, max_y)
    if not min_x or (min_x < 0) then
        min_x = 0
    end
    if not min_y or (min_y < 0) then
        min_y = 0
    end
    if not max_x or (max_x >= map.width) then
        max_x = map.width - 1
    end
    if not max_y or (max_y >= map.height) then
        max_y = map.height - 1
    end

    local layer = map.layers[index]
    if layer then
        for x = min_x, max_x do
            for y = min_y, max_y do
                draw_tile(map, index, x, y)
            end
        end
    end
end

local function draw_bottom(map, min_x, min_y, max_x, max_y)
    for i = 0, map.object_index do
        if map.layers[i] then
            draw_layer(map, i, min_x, min_y, max_x, max_y)
        end
    end
end
Map.draw_bottom = draw_bottom

local function draw_top(map, min_x, min_y, max_x, max_y)
    for i = map.object_index + 1, map.max_layer do
        if map.layers[i] then
            draw_layer(map, i, min_x, min_y, max_x, max_y)
        end
    end
end
Map.draw_top = draw_top

----------------------------------------------------------------------

return Map