--
-- Copyright 2012 Aaron MacDonald
--
-- Licensed under the Apache License, Version 2.0 (the "License");
-- you may not use this file except in compliance with the License.
-- You may obtain a copy of the License at
--
--     http://www.apache.org/licenses/LICENSE-2.0
--
-- Unless required by applicable law or agreed to in writing, software
-- distributed under the License is distributed on an "AS IS" BASIS,
-- WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
-- implied.
-- See the License for the specific language governing permissions and
-- limitations under the License.
--

local Menu = {}

local MouseManager = require 'mouse_manager'
local SoundManager = require 'sound_manager'

----------------------------------------------------------------------

local TEXT_COLOUR = {0, 255, 128, 255}
local HIGHLIGHT_COLOUR = {255, 255, 255, 255}

Menu.TEXT_COLOUR = TEXT_COLOUR
Menu.HIGHLIGHT_COLOUR = HIGHLIGHT_COLOUR

local SHADOW_COLOUR = {0, 128, 64, 255}
local SHADOW_HIGHLIGHT_COLOUR = {32, 160, 96, 255}

Menu.SHADOW_COLOUR = SHADOW_COLOUR
Menu.SHADOW_HIGHLIGHT_COLOUR = SHADOW_HIGHLIGHT_COLOUR

local SHADOW_SIZE = 3
Menu.SHADOW_SIZE = SHADOW_SIZE

local SELECT_SOUND = 'menu_select'
local CLICK_SOUND = 'menu_click'
Menu.CLICK_SOUND = CLICK_SOUND

local click_sound

----------------------------------------------------------------------

local function set_max_height(menu, max_height)
    menu.height = menu.line_height * #menu.choices

    if max_height and (menu.height > max_height) then
        local num_lines = math.floor(max_height / menu.line_height)

        -- Reserve spaces for scrollers
        menu.visible_lines = num_lines - 2

        menu.height = num_lines * menu.line_height
        menu.max_scroll = #menu.choices - menu.visible_lines
    else
        menu.visible_lines = #menu.choices
    end
end
Menu.set_max_height = set_max_height

local function recalculate_width(menu)
    local width
    for _, c in ipairs(menu.choices) do
        local w = menu.font:getWidth(c.text)
        if not width or (w > width) then
            width = w
        end
    end
    menu.width = width
end
Menu.recalculate_width = recalculate_width

local function new(x, y, font, choices, max_height)
    local line_height = font:getHeight() * font:getLineHeight()

    local result = {
        x = x,
        y = y,

        font = font,

        choices = choices,

        height = line_height * #choices,
        line_height = line_height,

        highlight_index = -1,
        scroll = 0,
        max_scroll = 0
    }

    recalculate_width(result)
    set_max_height(result, max_height)

    return result
end
Menu.new = new

----------------------------------------------------------------------

local function play_click_sound()
    click_sound = SoundManager.new_sound(CLICK_SOUND, nil, nil, nil,
      false, true)
end
Menu.play_click_sound = play_click_sound

local function play_select_sound()
    if not click_sound then
        SoundManager.new_sound(SELECT_SOUND, nil, nil, nil, false,
          true)
    end
end

local function run_option(menu, index)
    if not menu.choices[index].no_sound then
        play_click_sound()
    end
    menu.choices[index].func(index)
end
Menu.run_option = run_option

local function in_bounds(menu, x, y)
    local rx = x - menu.x
    local ry = y - menu.y

    return (rx >= 0) and (rx <= menu.width)
      and (ry >= 0) and (ry <= menu.height)
end

local function get_index(menu, y)
    local ry = y - menu.y
    return math.floor(ry / menu.line_height) + 1
end

local function update(menu)
    if click_sound and click_sound.source:isStopped() then
        click_sound = nil
    end

    local mx
    local my
    mx, my = MouseManager.get_pos()

    local new_index
    if in_bounds(menu, mx, my) then
        new_index = get_index(menu, my)
    else
        new_index = -1
    end

    if (new_index ~= -1) and (new_index ~= menu.highlight_index) then
        play_select_sound()
    end
    menu.highlight_index = new_index
end
Menu.update = update

local function scroll_up(menu)
    play_select_sound()
    menu.scroll = menu.scroll - 1
end
Menu.scroll_up = scroll_up

local function scroll_down(menu)
    play_select_sound()
    menu.scroll = menu.scroll + 1
end
Menu.scroll_down = scroll_down

local function click(menu, x, y)
    local result = false

    if in_bounds(menu, x, y) then
        result = true

        local index = get_index(menu, y)

        if menu.max_scroll > 0 then
            if (index == 1) and (menu.scroll > 0) then
                scroll_up(menu)
            elseif (index == (menu.visible_lines + 2))
              and (menu.scroll < menu.max_scroll) then
                scroll_down(menu)
            elseif (index > 1) and (index < (menu.visible_lines + 2))
              then
                run_option(menu, index + menu.scroll - 1)
            end
        else
            run_option(menu, index)
        end
    end

    return result
end
Menu.click = click

----------------------------------------------------------------------

local function draw_shadowed_text(text, x, y, colour, shadow_colour)
    if not colour then
        colour = TEXT_COLOUR
    end
    if not shadow_colour then
        shadow_colour = SHADOW_COLOUR
    end

    love.graphics.setColor(shadow_colour)
    for i = 1, SHADOW_SIZE do
        love.graphics.print(text,
          x + i, y + i)
    end

    love.graphics.setColor(colour)
    love.graphics.print(text, x, y)

    love.graphics.setColor(255, 255, 255, 255)
end
Menu.draw_shadowed_text = draw_shadowed_text

local function draw_text(menu, text, line)
    local lx = ((menu.width - menu.font:getWidth(text)) / 2)
    local ly = (menu.line_height * (line - 1))

    local colour
    local shadow_colour
    if line == menu.highlight_index then
        colour = HIGHLIGHT_COLOUR
        shadow_colour = SHADOW_HIGHLIGHT_COLOUR
    else
        colour = TEXT_COLOUR
        shadow_colour = SHADOW_COLOUR
    end

    draw_shadowed_text(text, menu.x + lx, menu.y + ly,
      colour, shadow_colour)
end

local function draw(menu)
    love.graphics.setFont(menu.font)

    local line_add = 0
    if menu.max_scroll > 0 then
        line_add = 1
        if menu.scroll > 0 then
            draw_text(menu, '- Up -', 1)
        end
        if menu.scroll < menu.max_scroll then
            draw_text(menu, '- Down -', menu.visible_lines + 2)
        end
    end

    for i = 1, menu.visible_lines do
        local text = menu.choices[i + menu.scroll].text
        draw_text(menu, text, i + line_add)
    end

    love.graphics.setColor(255, 255, 255, 255)
end
Menu.draw = draw

----------------------------------------------------------------------

return Menu