--
-- Copyright 2012 Aaron MacDonald
--
-- Licensed under the Apache License, Version 2.0 (the "License");
-- you may not use this file except in compliance with the License.
-- You may obtain a copy of the License at
--
--     http://www.apache.org/licenses/LICENSE-2.0
--
-- Unless required by applicable law or agreed to in writing, software
-- distributed under the License is distributed on an "AS IS" BASIS,
-- WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
-- implied.
-- See the License for the specific language governing permissions and
-- limitations under the License.
--

local SoundManager = {}

local Constants = require 'constants'
local Vector = require 'vector'
local Settings = require 'settings'

local MAX_DISTANCE = 1024

local static_sound_data

local current_sounds
local sound_pool

local listener_pos

----------------------------------------------------------------------

local function load()
    static_sound_data = {}
    current_sounds = {}
    sound_pool = {}

    listener_pos = Vector.new()
end
SoundManager.load = load

local function set_listener(x, y)
    listener_pos.x = x
    listener_pos.y = y
end
SoundManager.set_listener = set_listener

local function reset_listener()
    listener_pos.x = 0
    listener_pos.y = 0
end
SoundManager.reset_listener = reset_listener

-- unclearable = Not affected by SoundManager.clear
-- persist = Not removed even if stopped
local function new_sound(file_name, x, y, max_hearing_dist,
  loops, unclearable, start_stopped, persist)
    local new_sound = {file_name = file_name}
    local volume = 1

    if x then
        assert(y)
        new_sound.pos = Vector.new(x, y)

        local max_dist = MAX_DISTANCE
        if max_hearing_dist then
            max_dist = max_hearing_dist
        end
        new_sound.max_dist = max_dist

        local dist = Vector.distance(listener_pos, Vector.new(x, y))
        volume = math.max(0, 1 - (dist / max_dist))
    end

    if not static_sound_data[file_name] then
        static_sound_data[file_name] = love.sound.newSoundData(
            Constants.SOUND_PATH .. '/' .. file_name .. '.'
            .. Constants.DEFAULT_SOUND_TYPE
        )
    end

    if sound_pool[file_name] and (#sound_pool[file_name] > 0) then
        new_sound.source = table.remove(sound_pool[file_name])
    else
        new_sound.source
          = love.audio.newSource(static_sound_data[file_name])
    end

    new_sound.source:setVolume(volume * Settings.get('volume'))
    new_sound.source:setLooping(loops)

    if not start_stopped then
        new_sound.source:play()
    end

    new_sound.unclearable = unclearable
    new_sound.persist = persist

    table.insert(current_sounds, new_sound)

    return new_sound
end
SoundManager.new_sound = new_sound

local function clear()
    for _, s in ipairs(current_sounds) do
        if not s.unclearable then
            s.source:stop()
            if not sound_pool[s.file_name] then
                sound_pool[s.file_name] = {}
            end
            table.insert(sound_pool[s.file_name], s.source)
        end
    end
    current_sounds = {}
end
SoundManager.clear = clear

local function pause_current()
    for _, s in ipairs(current_sounds) do
        s.source:pause()
    end
end
SoundManager.pause_current = pause_current

local function resume()
    for _, s in ipairs(current_sounds) do
        s.source:resume()
    end
end
SoundManager.resume = resume

local function update()
    local i = 1
    while i <= #current_sounds do
        local s = current_sounds[i]

        if s.source:isStopped() and not s.persist then
            if not sound_pool[s.file_name] then
                sound_pool[s.file_name] = {}
            end
            table.insert(sound_pool[s.file_name], s.source)

            table.remove(current_sounds, i)
        else
            if s.pos then
                local max_dist = MAX_DISTANCE
                if s.max_dist then
                    max_dist = s.max_dist
                end

                local dist = Vector.distance(listener_pos, s.pos)
                local volume = math.max(0, 1 - (dist / max_dist))
                s.source:setVolume(volume * Settings.get('volume'))
            end

            i = i + 1
        end
    end
end
SoundManager.update = update

----------------------------------------------------------------------

return SoundManager