--
-- Copyright 2012 Aaron MacDonald
--
-- Licensed under the Apache License, Version 2.0 (the "License");
-- you may not use this file except in compliance with the License.
-- You may obtain a copy of the License at
--
--     http://www.apache.org/licenses/LICENSE-2.0
--
-- Unless required by applicable law or agreed to in writing, software
-- distributed under the License is distributed on an "AS IS" BASIS,
-- WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
-- implied.
-- See the License for the specific language governing permissions and
-- limitations under the License.
--

local Spawn = {}

local Constants = require 'constants'
local Utility = require 'utility'

local Actor = require 'actor'
local Particles = require 'particles'

local SoundManager = require 'sound_manager'

----------------------------------------------------------------------

local SPAWN_IMG = 'spawn'

local NEW_SPAWN_SOUND = 'new_spawn'
local WORKING_SOUND = 'spawn_working'
local RESURRECT_SOUND = 'resurrect'

local SPAWN_WIDTH = 99

local TRANS_X = SPAWN_WIDTH / 2

local SPAWN_TOP_BOTTOM_HEIGHT = SPAWN_WIDTH
local SPAWN_FRONT_BACK_HEIGHT = 85

local GLOW_BORDER = 4

local BOTTOM_TRANS_Y = SPAWN_TOP_BOTTOM_HEIGHT / 2
local BACK_TRANS_Y = SPAWN_FRONT_BACK_HEIGHT - GLOW_BORDER
local FRONT_TRANS_Y = (SPAWN_FRONT_BACK_HEIGHT / 2) - GLOW_BORDER
local TOP_TRANS_Y = SPAWN_TOP_BOTTOM_HEIGHT - (GLOW_BORDER * 4)

local SPAWN_TIME = 300

local spawn_image

local bottom_top_quad
local back_quad
local front_quad

----------------------------------------------------------------------

local function new(map, section, pos)
    if not box_image then
        spawn_image = love.graphics.newImage(
          Utility.img_path(SPAWN_IMG))

        local img_width = spawn_image:getWidth()
        local img_height = spawn_image:getHeight()

        bottom_top_quad = love.graphics.newQuad(
          0, 0, SPAWN_WIDTH, SPAWN_TOP_BOTTOM_HEIGHT,
          img_width, img_height
        )

        back_quad = love.graphics.newQuad(
          SPAWN_WIDTH, 0, SPAWN_WIDTH, SPAWN_FRONT_BACK_HEIGHT,
          img_width, img_height
        )

        front_quad = love.graphics.newQuad(
          SPAWN_WIDTH * 2, 0, SPAWN_WIDTH, SPAWN_FRONT_BACK_HEIGHT,
          img_width, img_height
        )
    end

    local particle_source = Particles.create_fountain(
        pos.x, pos.y,
        1000, 500,
        96,
        100, 200
    )
    particle_source:setColors(
          0, 255, 128, 255,
        255, 255, 255, 128
    )

    return {
        map = map,
        section = section,
        pos = pos,

        spawn_time = 0,
        queue = {},

        particle_source = particle_source,
        working_sound = SoundManager.new_sound(
            WORKING_SOUND, pos.x, pos.y, nil,
            true, false, true, true
        )
    }
end
Spawn.new = new

local function schedule_spawn(spawn, actor_name)
    if #spawn.queue == 0 then
        spawn.spawn_time = SPAWN_TIME
        spawn.particle_source:start()
    end
    table.insert(spawn.queue, actor_name)

    spawn.working_sound.source:play()
    SoundManager.new_sound(NEW_SPAWN_SOUND, spawn.pos.x, spawn.pos.y)
end
Spawn.schedule_spawn = schedule_spawn

local function update(spawn)
    Particles.update(spawn.particle_source)

    local new_monster
    local spawn_explosion

    if spawn.spawn_time > 0 then
        spawn.spawn_time = spawn.spawn_time - 1
        if spawn.spawn_time == 0 then
            assert(#spawn.queue > 0)

            local monster_name = spawn.queue[1]
            table.remove(spawn.queue, 1)

            local shift_x = 1
            if math.random() < 0.5 then
                shift_x = -1
            end

            local shift_y = 1
            if math.random() < 0.5 then
                shift_y = -1
            end

            local mon_x = spawn.pos.x
              + (Constants.TILE_HALF_SIZE * shift_x)
            local mon_y = spawn.pos.y
              + (Constants.TILE_HALF_SIZE * shift_y)

            new_monster = Actor.new(
                monster_name, spawn.map, mon_x, mon_y
            )
            new_monster.ai_data = {
                section = spawn.section,
                wandering = true
            }

            spawn_explosion = Particles.create_explosion(
              mon_x, mon_y, 96, 96, 45)
            spawn_explosion:setColors(
                  0, 255, 128, 255,
                255, 255, 255, 128
            )

            SoundManager.new_sound(RESURRECT_SOUND,
              spawn.pos.x, spawn.pos.y)

            if #spawn.queue > 0 then
                spawn.spawn_time = SPAWN_TIME
            else
                spawn.working_sound.source:stop()
                spawn.particle_source:stop()
            end
        end
    end

    return new_monster, spawn_explosion
end
Spawn.update = update

local function destroy(spawn)
    spawn.working_sound.source:stop()
    spawn.working_sound.persist = false
    spawn.particle_source:stop()
end
Spawn.destroy = destroy

local function draw_data(spawn)
    return {
        { -- Bottom
            img = spawn_image,
            quad = bottom_top_quad,
            x = spawn.pos.x - TRANS_X,
            y = spawn.pos.y - BOTTOM_TRANS_Y,

            real_y = spawn.pos.y + BOTTOM_TRANS_Y,
            z = 0
        },
        { -- Back
            img = spawn_image,
            quad = back_quad,
            x = spawn.pos.x - TRANS_X,
            y = spawn.pos.y - BACK_TRANS_Y,

            real_y = spawn.pos.y - (BACK_TRANS_Y / 2),
            z = 1
        },
        { -- Front
            img = spawn_image,
            quad = front_quad,
            x = spawn.pos.x - TRANS_X,
            y = spawn.pos.y - FRONT_TRANS_Y,

            real_y = spawn.pos.y + FRONT_TRANS_Y,
            z = 1
        },
        { -- Top
            img = spawn_image,
            quad = bottom_top_quad,
            x = spawn.pos.x - TRANS_X,
            y = spawn.pos.y - TOP_TRANS_Y,

            real_y = spawn.pos.y + (GLOW_BORDER * 2),
            z = 2
        }
    }
end
Spawn.draw_data = draw_data

local function draw_particles(spawn)
    Particles.draw(spawn.particle_source)
end
Spawn.draw_particles = draw_particles

----------------------------------------------------------------------

return Spawn