--
-- Copyright 2012 Aaron MacDonald
--
-- Licensed under the Apache License, Version 2.0 (the "License");
-- you may not use this file except in compliance with the License.
-- You may obtain a copy of the License at
--
--     http://www.apache.org/licenses/LICENSE-2.0
--
-- Unless required by applicable law or agreed to in writing, software
-- distributed under the License is distributed on an "AS IS" BASIS,
-- WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
-- implied.
-- See the License for the specific language governing permissions and
-- limitations under the License.
--

local Title = {}

local Constants = require 'constants'
local Utility = require 'utility'

local Settings = require 'settings'

local ScreenManager = require 'screen_manager'
local SoundManager = require 'sound_manager'

local ColourFlux = require 'colour_flux'
local Menu = require 'menu'

----------------------------------------------------------------------

local start_new_game

local LOGO_NAME = 'logo'
local LOGO_Y_POS = 0.05 -- Percent

local logo_img

local COLOURS = {
    {255,   0,   0, 255}, -- Red
    {255, 128,   0, 255}, -- Orange
    {255, 255,   0, 255}, -- Yellow
    {  0, 255,   0, 255}, -- Green
    {  0, 255, 255, 255}, -- Cyan
    {  0,   0, 255, 255}, -- Blue
    {255,   0, 255, 255}  -- Violet
}
local COLOUR_SPEED = 0.1

local colour_flux

local TEXT_NAME_SIZE = 16
local TEXT_MENU_SIZE = 32

local SET_VOLUME_TEXT = 'Volume'

local VOLUME_BAR_HEIGHT = 32
local VOLUME_BAR_WIDTH_PERCENT = 0.667

local VOLUME_CHANGE = 0.05

local name_font
local menu_font

local TEXT_NAME = 'Aaron MacDonald, 2012'
local text_name_width

local current_menu

local main_menu
local options_menu
local screen_menu
local key_menu

local volume_back_menu
local key_cancel_menu

local key_prompt
local volume_prompt

----------------------------------------------------------------------

local MAIN_MENU_CHOICES = {
    {
        text = 'New Game',
        func = function()
            start_new_game = true
        end
    },
    {
        text = 'Options',
        func = function()
            current_menu = options_menu
        end
    },
    {
        text = 'Quit',
        no_sound = true, -- Sound will only get cut off when game
                         -- closes
        func = function()
            love.event.push('quit')
        end
    }
}

local OPTIONS_MENU_CHOICES = {
    {
        text = 'Set Resolution',
        func = function()
            current_menu = screen_menu
        end
    },
    {
        text = 'Toggle Fullscreen',
        func = function()
            love.graphics.toggleFullscreen()
            local is_fullscreen
            _, _, is_fullscreen, _, _ = love.graphics.getMode()

            Settings.set('fullscreen', is_fullscreen)
        end
    },
    {
        text = 'Set Keys',
        func = function()
            current_menu = key_menu
        end
    },
    {
        text = 'Set Volume',
        func = function()
            volume_prompt.active = true
            volume_prompt.volume = Settings.get('volume')
        end
    },
    {
        text = 'Back',
        func = function()
            current_menu = main_menu
        end
    }
}

local SET_KEY_CHOICES_STRUCTURE = {
    key_north = {
        index = 1,
        text = 'Move North',
        prompt = 'moving north'
    },
    key_east = {
        index = 2,
        text = 'Move East',
        prompt = 'moving east'
    },
    key_south = {
        index = 3,
        text = 'Move South',
        prompt = 'moving south'
    },
    key_west = {
        index = 4,
        text = 'Move West',
        prompt = 'moving west'
    },
    key_mystery_box = {
        index = 5,
        text = 'Use Mystery Box',
        prompt = 'using a mystery box'
    },
    key_screenshot = {
        index = 6,
        text = 'Take Screenshot',
        prompt = 'taking a screenshot'
    }
}
local set_key_choices

local VOLUME_BACK_MENU_CHOICE = {
    {
        text = 'Back',
        func = function()
            volume_prompt.active = false
        end
    }
}

local KEY_CANCEL_MENU_CHOICE = {
    {
        text = 'Cancel',
        func = function()
            key_prompt.active = false
        end
    }
}

----------------------------------------------------------------------

local function activate_key_prompt(description, key)
    key_prompt.active = true

    key_prompt.text = 'Press key for ' .. description
    key_prompt.width = menu_font:getWidth(key_prompt.text)

    key_prompt.key = key
end

local function change_key(key_control, key_value)
    Settings.set(key_control, key_value)

    local key_data = SET_KEY_CHOICES_STRUCTURE[key_control]
    set_key_choices[key_data.index].text = key_data.text .. ': '
      .. Settings.get_key_control_name(key_control)
end

local function set_key_prompt(new_key)
    local old_key = Settings.get(key_prompt.key)

    for _, k in ipairs(Settings.KEY_CONTROLS) do
        if k ~= key_prompt.key then
            if Settings.get(k) == new_key then
                change_key(k, old_key)
                break -- Assume we don't start with duplicates
            end
        end
    end
    change_key(key_prompt.key, new_key)

    Menu.recalculate_width(key_menu)
    key_prompt.active = false
end

local function change_volume(new_volume)
    volume_prompt.volume = new_volume
    Settings.set('volume', volume_prompt.volume)
    SoundManager.new_sound(Menu.CLICK_SOUND)
end

----------------------------------------------------------------------

local function load()
    colour_flux = ColourFlux.new(COLOURS, COLOUR_SPEED)

    -- Logo

    logo_img = love.graphics.newImage(Utility.img_path(LOGO_NAME))

    -- Name text

    name_font = love.graphics.newFont(Constants.DEFAULT_FONT,
      TEXT_NAME_SIZE)
    text_name_width = name_font:getWidth(TEXT_NAME)

    -- Menus

    menu_font = love.graphics.newFont(Constants.DEFAULT_FONT,
      TEXT_MENU_SIZE)
    menu_font:setLineHeight(1.5)

    main_menu = Menu.new(0, 0, menu_font, MAIN_MENU_CHOICES)
    options_menu = Menu.new(0, 0, menu_font, OPTIONS_MENU_CHOICES)

    -- Resolution menu

    local modes = love.graphics.getModes()
    table.sort(modes, function(a, b)
        return (a.width * a.height) < (b.width * b.height)
    end)

    local screen_choices = {}
    for _, m in ipairs(modes) do
        table.insert(screen_choices, {
            text = m.width .. ' x ' .. m.height,
            func = function()
                Settings.set('screen_width', m.width)
                Settings.set('screen_height', m.height)

                love.graphics.setMode(m.width, m.height,
                  Settings.get('fullscreen'))
            end
        })
    end
    table.insert(screen_choices, {
        text = 'Back',
        func = function()
            current_menu = options_menu
        end
    })

    screen_menu = Menu.new(0, 0, menu_font, screen_choices)

    -- Set Key menu

    set_key_choices = {}
    for key, data in pairs(SET_KEY_CHOICES_STRUCTURE) do
        set_key_choices[data.index] = {
            text = data.text .. ': '
              .. Settings.get_key_control_name(key),
            func = function()
                activate_key_prompt(data.prompt, key)
            end
        }
    end
    table.insert(set_key_choices, {
        text = 'Back',
        func = function()
            current_menu = main_menu
        end
    })

    key_menu = Menu.new(0, 0, menu_font, set_key_choices)

    -- Back "button" for volume control
    volume_back_menu = Menu.new(0, 0, menu_font,
      VOLUME_BACK_MENU_CHOICE)

    -- Cancel "button" for key configuration
    key_cancel_menu = Menu.new(0, 0, menu_font,
      KEY_CANCEL_MENU_CHOICE)

    --

    current_menu = main_menu
    key_prompt = {active = false}
    volume_prompt = {active = false}
end
Title.load = load

local function init()
    start_new_game = false
end
Title.init = init

local function close_title()
    ColourFlux.reset(colour_flux)
end

----------------------------------------------------------------------

local function get_header_positions(screen_width, screen_height)
    local logo_x = (screen_width - logo_img:getWidth()) / 2
    local logo_y = screen_height * LOGO_Y_POS

    local text_name_x = (screen_width - text_name_width) / 2
    local text_name_y = logo_y + logo_img:getHeight()

    return logo_x, logo_y, text_name_x, text_name_y
end

local function get_key_prompt_positions(screen_width, screen_height)
    local top_y
    _, _, _, top_y = get_header_positions(screen_width, screen_height)

    local text_height
      = menu_font:getHeight() * menu_font:getLineHeight()

    local total_height = text_height + key_cancel_menu.height

    local prompt_x = (screen_width - key_prompt.width) / 2
    local prompt_y = ((screen_height - top_y - total_height) / 2)
      + top_y

    local cancel_x = (screen_width - key_cancel_menu.width) / 2
    local cancel_y = prompt_y + text_height

    return prompt_x, prompt_y, cancel_x, cancel_y
end

local function get_volume_bar_positions(screen_width, screen_height)
    local top_y
    _, _, _, top_y = get_header_positions(screen_width, screen_height)

    local text_height
      = menu_font:getHeight() * menu_font:getLineHeight()
    local border_height = text_height - menu_font:getHeight()

    local total_height = text_height + VOLUME_BAR_HEIGHT
      + volume_back_menu.height

    local text_width = menu_font:getWidth(SET_VOLUME_TEXT)

    local text_x = (screen_width - text_width) / 2
    local text_y = ((screen_height - top_y - total_height) / 2)
      + top_y

    local bar_total_width
      = screen_width * VOLUME_BAR_WIDTH_PERCENT

    local bar_filled_width = bar_total_width * volume_prompt.volume
    local bar_empty_width = bar_total_width - bar_filled_width

    local bar_x = (screen_width - bar_total_width) / 2
    local bar_y = text_x + text_height

    local back_x = (screen_width - volume_back_menu.width) / 2
    local back_y = bar_y + VOLUME_BAR_HEIGHT + border_height

    return text_x, text_y, bar_x, bar_y, back_x, back_y,
      bar_total_width, bar_filled_width, bar_empty_width
end

local function position_current_menu(screen_width, screen_height)
    local top_y
    _, _, _, top_y = get_header_positions(screen_width, screen_height)

    local prompt_y = 

    Menu.set_max_height(current_menu, screen_height - top_y)

    current_menu.x = (screen_width - current_menu.width) / 2
    current_menu.y = ((screen_height - top_y - current_menu.height)
      / 2) + top_y
end

----------------------------------------------------------------------

local function mouse_data()
    return 'mouse', 0, 0
end
Title.mouse_data = mouse_data

local function mouse_press(x, y)
    local next_state

    local screen_width
    local screen_height
    screen_width, screen_height = ScreenManager.dimensions()

    local text_name_y
    _, _, _, text_name_y = get_header_positions(
      screen_width, screen_height)

    if key_prompt.active then
        local cancel_x
        local cancel_y

        _, _, cancel_x, cancel_y = get_key_prompt_positions(
          screen_width, screen_height)

        key_cancel_menu.x = cancel_x
        key_cancel_menu.y = cancel_y

        Menu.click(key_cancel_menu, x, y)
    elseif volume_prompt.active then
        local bar_x
        local bar_y
        local back_x
        local back_y
        local bar_total_width

        _, _, bar_x, bar_y, back_x, back_y, bar_total_width, _, _
          = get_volume_bar_positions(screen_width, screen_height)

        volume_back_menu.x = back_x
        volume_back_menu.y = back_y

        if not Menu.click(volume_back_menu, x, y) then
            if x < bar_x then
                change_volume(0)
            elseif x > (bar_x + bar_total_width) then
                change_volume(1)
            else
                local click_x = x - bar_x
                change_volume(click_x / bar_total_width)
            end
        end
    else
        position_current_menu(screen_width, screen_height)
        Menu.click(current_menu, x, y)

        if start_new_game then
            close_title()
            next_state = Constants.STATE_LEVEL_WARP
        end
    end

    return next_state
end
Title.mouse_press = mouse_press

local function mouse_release()
end
Title.mouse_release = mouse_release

local function wheel_up()
    if volume_prompt.active and (volume_prompt.volume < 1) then
        change_volume(math.min(
            1, volume_prompt.volume + VOLUME_CHANGE
        ))
    elseif (current_menu.max_scroll > 0) and (current_menu.scroll > 0)
      then
        Menu.scroll_up(current_menu)
    end
end
Title.wheel_up = wheel_up

local function wheel_down()
    if volume_prompt.active and (volume_prompt.volume > 0) then
        change_volume(math.max(
            0, volume_prompt.volume - VOLUME_CHANGE
        ))
    elseif (current_menu.max_scroll > 0)
      and (current_menu.scroll < current_menu.max_scroll) then
        Menu.scroll_down(current_menu)
    end
end
Title.wheel_down = wheel_down

local function keypressed(key, unicode)
    if key == 'escape' then
        if key_prompt.active then
            key_prompt.active = false
        elseif volume_prompt.active then
            volume_prompt.active = false
        else
            -- All menues have a 'Back' option
            Menu.run_option(current_menu, #current_menu.choices)
        end
    elseif (key == Settings.get('key_screenshot'))
      and not key_prompt.active then
        Utility.take_screenshot()
    elseif key_prompt.active then
        set_key_prompt(key)
    end
end
Title.keypressed = keypressed

local function keyreleased(key, unicode)
end
Title.keyreleased = keyreleased

local function set_paused(is_paused)
end
Title.set_paused = set_paused

----------------------------------------------------------------------

local function update(key, unicode)
    ColourFlux.update(colour_flux)

    local screen_width
    local screen_height
    screen_width, screen_height = ScreenManager.dimensions()

    if key_prompt.active then
        local cancel_x
        local cancel_y

        _, _, cancel_x, cancel_y = get_key_prompt_positions(
          screen_width, screen_height)

        key_cancel_menu.x = cancel_x
        key_cancel_menu.y = cancel_y

        Menu.update(key_cancel_menu)
    elseif volume_prompt.active then
        local back_x
        local back_y

        _, _, _, _, back_x, back_y, _, _, _
          = get_volume_bar_positions(screen_width, screen_height)

        volume_back_menu.x = back_x
        volume_back_menu.y = back_y

        Menu.update(volume_back_menu)
    else
        position_current_menu(screen_width, screen_height)
        Menu.update(current_menu)
    end
end
Title.update = update

----------------------------------------------------------------------

local function draw()
    local screen_width
    local screen_height
    screen_width, screen_height = ScreenManager.dimensions()

    local logo_x
    local logo_y
    local text_name_x
    local text_name_y

    logo_x, logo_y, text_name_x, text_name_y
      = get_header_positions(screen_width, screen_height)

    love.graphics.setColor(ColourFlux.get_colour(colour_flux))
    love.graphics.draw(logo_img, logo_x, logo_y)

    love.graphics.setColor(255, 255, 255)

    love.graphics.setFont(name_font)
    love.graphics.print(TEXT_NAME, text_name_x, text_name_y)

    if key_prompt.active then
        love.graphics.setFont(menu_font)

        local prompt_x
        local prompt_y
        local cancel_x
        local cancel_y

        prompt_x, prompt_y, cancel_x, cancel_y
          = get_key_prompt_positions(screen_width, screen_height)

        Menu.draw_shadowed_text(key_prompt.text, prompt_x, prompt_y)

        key_cancel_menu.x = cancel_x
        key_cancel_menu.y = cancel_y

        Menu.draw(key_cancel_menu)
    elseif volume_prompt.active then
        love.graphics.setFont(menu_font)

        local text_x
        local text_y
        local bar_x
        local bar_y
        local back_x
        local back_y
        local bar_filled_width
        local bar_empty_width

        text_x, text_y, bar_x, bar_y, back_x, back_y,
          _, bar_filled_width, bar_empty_width
          = get_volume_bar_positions(screen_width, screen_height)

        Menu.draw_shadowed_text(SET_VOLUME_TEXT, text_x, text_y)

        love.graphics.setColor(Menu.TEXT_COLOUR)
        love.graphics.rectangle(
            'fill', bar_x, bar_y, bar_filled_width, VOLUME_BAR_HEIGHT
        )
        love.graphics.setColor(Menu.SHADOW_COLOUR)
        love.graphics.rectangle(
            'fill', bar_x + bar_filled_width, bar_y,
            bar_empty_width, VOLUME_BAR_HEIGHT
        )

        volume_back_menu.x = back_x
        volume_back_menu.y = back_y

        Menu.draw(volume_back_menu)
    else
        position_current_menu(screen_width, screen_height)
        Menu.draw(current_menu)
    end
end
Title.draw = draw

----------------------------------------------------------------------

return Title