--
-- Copyright 2012 Aaron MacDonald
--
-- Licensed under the Apache License, Version 2.0 (the "License");
-- you may not use this file except in compliance with the License.
-- You may obtain a copy of the License at
--
--     http://www.apache.org/licenses/LICENSE-2.0
--
-- Unless required by applicable law or agreed to in writing, software
-- distributed under the License is distributed on an "AS IS" BASIS,
-- WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
-- implied.
-- See the License for the specific language governing permissions and
-- limitations under the License.
--

local Utility = {}

local Constants = require 'constants'

----------------------------------------------------------------------

local function find_in_array(arr, val, equ)
    local result_v = nil
    local result_i = nil

    if not equ then
        equ = function (a, b) return a == b end
    end

    for i, v in ipairs(arr) do
        if equ(v, val) then
            result_v = v
            result_i = i
            break
        end
    end

    return result_v, result_i
end
Utility.find_in_array = find_in_array

local function find_key(tab, val, equ)
    local result = nil

    if not equ then
        equ = function (a, b) return a == b end
    end

    for k, _ in pairs(tab) do
        if equ(k, val) then
            result = k
            break
        end
    end

    return result
end
Utility.find_key = find_key

local function keys_to_array(tab, sort_f)
    local result = {}
    for k, _ in pairs(tab) do
        result[#result + 1] = k
    end

    if sort_f then
        table.sort(result, sort_f)
    end

    return result
end
Utility.keys_to_array = keys_to_array

local function remove_from_array(arr, val, equ)
    local result = false
    local _, i = find_in_array(arr, val, equ)
    if i then
        table.remove(arr, i)
        result = true
    end
    return result
end
Utility.remove_from_array = remove_from_array

local function round(num)
    local result
    if num >= 0 then
        result = math.floor(num + 0.5)
    else
        result = math.ceil(num + 0.5)
    end

    return result
end
Utility.round = round

local function img_path(name)
    return Constants.GRAPHICS_PATH .. '/' .. name .. '.'
      .. Constants.DEFAULT_IMG_TYPE
end
Utility.img_path = img_path

local function set_double_key(tab, key1, key2, val)
    if not tab[key1] then
        tab[key1] = {}
    end
    tab[key1][key2] = val
end
Utility.set_double_key = set_double_key

local function new_tab_insert(tab, key, val)
    if not tab[key] then
        tab[key] = {}
    end
    table.insert(tab[key], val)
end
Utility.new_tab_insert = new_tab_insert

-- interpolate(a, b, 0) = a
-- interpolate(a, b, 1) = b
local function interpolate(a, b, delta)
    return (a * (1 - delta)) + (b * delta)
end
Utility.interpolate = interpolate

local function interpolate_colour(
  r1, g1, b1, a1,
  r2, g2, b2, a2,
  delta)
    local r = interpolate(r1, r2, delta)
    local g = interpolate(g1, g2, delta)
    local b = interpolate(b1, b2, delta)
    local a = interpolate(a1, a2, delta)

    return r, g, b, a
end
Utility.interpolate_colour = interpolate_colour

local function coords_to_angle(x, y)
    local angle

    if x > 0 then
        angle = math.atan(y / x)
    elseif (x < 0) and (y >= 0) then
        angle = math.atan(y / x) + math.pi
    elseif (x < 0) and (y < 0) then
        angle = math.atan(y / x) - math.pi
    elseif (x == 0) and (y > 0) then
        angle = math.pi / 2
    elseif (x == 0) and (y < 0) then
        angle = -math.pi / 2
    else
        assert(x == 0)
        assert(y == 0)
        angle = 0
    end

    return angle
end
Utility.coords_to_angle = coords_to_angle

local function take_screenshot()
    local screenshot = love.graphics.newScreenshot()

    local time_string = os.date('%Y-%m-%d_%H-%M-%S')
    local filename = 'warp_run_' .. time_string .. '.'
      .. Constants.DEFAULT_IMG_TYPE

    if not love.filesystem.exists('screenshots')
      or not love.filesystem.isDirectory('screenshots') then
        love.filesystem.mkdir('screenshots')
    end

    screenshot:encode('screenshots/' .. filename,
      Constants.DEFAULT_IMG_TYPE)
end
Utility.take_screenshot = take_screenshot

----------------------------------------------------------------------

return Utility