--
-- Copyright 2012 Aaron MacDonald
--
-- Licensed under the Apache License, Version 2.0 (the "License");
-- you may not use this file except in compliance with the License.
-- You may obtain a copy of the License at
--
--     http://www.apache.org/licenses/LICENSE-2.0
--
-- Unless required by applicable law or agreed to in writing, software
-- distributed under the License is distributed on an "AS IS" BASIS,
-- WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
-- implied.
-- See the License for the specific language governing permissions and
-- limitations under the License.
--

local Vector = {}

----------------------------------------------------------------------

local function new(x, y)
    if not x then
        x = 0
    end
    if not y then
        y = 0
    end
    return {x = x, y = y}
end
Vector.new = new

local function to_string(p)
    return '(' .. p.x .. ', ' .. p.y .. ')'
end
Vector.to_string = to_string

local function equals(a, b)
    return (a.x == b.x) and (a.y == b.y)
end
Vector.equals = equals

local function add(a, b)
    return new(a.x + b.x, a.y + b.y)
end
Vector.add = add

local function subtract(a, b)
    return new(a.x - b.x, a.y - b.y)
end
Vector.subtract = subtract

local function scale(p, s)
    return new(p.x * s, p.y * s)
end
Vector.scale = scale

local function scale_inline(p, s)
    p.x = p.x * s
    p.y = p.y * s
end
Vector.scale_inline = scale_inline

local function length_sqr(p)
    return (p.x * p.x) + (p.y * p.y)
end
Vector.length_sqr = length_sqr

local function length(p)
    return math.sqrt(length_sqr(p))
end
Vector.length = length

local function distance(start_v, end_v)
    return length(subtract(end_v, start_v))
end
Vector.distance = distance

local function distance_sqr(start_v, end_v)
    return length_sqr(subtract(end_v, start_v))
end
Vector.distance_sqr = distance_sqr

local function normal(v)
    local d = length(v)
    return new(v.x / d, v.y / d)
end
Vector.normal = normal

----------------------------------------------------------------------

local DIRS = {
    north     = new( 0, -1),
    northeast = new( 1, -1),
    east      = new( 1,  0),
    southeast = new( 1,  1),
    south     = new( 0,  1),
    southwest = new(-1,  1),
    west      = new(-1,  0),
    northwest = new(-1, -1)
}
Vector.DIRS = DIRS

local CARDINAL_DIRS = {
    north = DIRS.north,
    east = DIRS.east,
    south = DIRS.south,
    west = DIRS.west
}
Vector.CARDINAL_DIRS = CARDINAL_DIRS

local function get_adjacent(all_dirs, p)
    local result = {}
    for _, delta in pairs(all_dirs) do
        result[#result + 1] = add(p, delta)
    end

    return result
end

local function adjacent(p)
    return get_adjacent(DIRS, p)
end
Vector.adjacent = adjacent

local function cardinal_adjacent(p)
    return get_adjacent(CARDINAL_DIRS, p)
end
Vector.cardinal_adjacent = cardinal_adjacent

local function get_dir(start_p, end_p)
    assert(not Vector.equals(start_p, end_p))

    local result_name = nil
    local result_delta = nil

    local delta = subtract(end_p, start_p)

    local div_x = math.abs(delta.x)
    if div_x == 0 then
        div_x = 1
    end
    local div_y = math.abs(delta.y)
    if div_y == 0 then
        div_y = 1
    end

    delta.x = delta.x / div_x
    delta.y = delta.y / div_y

    for dir_name, dir_delta in pairs(DIRS) do
        if equals(delta, dir_delta) then
            result_name = dir_name
            result_delta = delta
            break
        end
    end

    return result_name, result_delta
end
Vector.get_dir = get_dir

----------------------------------------------------------------------

return Vector